Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_arith.all;
USE ieee.std_logic_unsigned.all;
USE work.pkg_081006.all; 

--------------------------------------------------------------
-- Description:
-- This module generates two input signals for one or more
-- AD7625, 16-bit Analog to Digital Converters (ADC).  The
-- signal adc_cnv signals the ADC to start a conversion, and
-- the signal adc_clk is a clock for serially clocking data
-- out of the ADC.  See the AD7625 datasheet for full details
-- of the ADC's digital interface (rev0|Page 19 of 24).
--
-- The timing in this logic reads data from the ADC at a 5 MHz
-- rate.  A new conversion is signaled approximately mid-way 
-- through reading out the previous conversion.  The data clock
-- is provided at 100 MHz.  This is produced with a PLL that
-- multiplies the 50 MHz logic clock by 2.  The data clock must
-- be gated on and off with the specified timing w.r.t. the 
-- convert pulse.  The timing is accomplished using a counter
-- (timer) and a state machine.
--
-- Gating the data clock has some details.  The counter/timer
-- and state machine, which produces the data clock gate, is
-- syncronized to the global Logic Clock (50 MHz).  This gate
-- is syncronized to the PLL output clock (100 MHz) before 
-- being used to gate the PLL output clock as the data clock.
--
-- NOTE: Timing requires that there be a minimum of 145ns between
-- the rising edge of adc_cnv and the first rising edge of the 
-- gated data clock.  Also the last falling edge of the gated
-- data clock must occur within 110ns after the rising edge of the
-- next adc_cnv pulse.
-- 
-- If the gated data clock burst needs to be shifted +/-10 ns
-- consider changing the number of registers used to syncronize
-- the gate to the PLL output clock. 

entity AD7625_INTERFACE_v2 is
	port(clk       : in std_logic; -- global Logic Clock
		 reset_n   : in std_logic; -- global Logic Reset
		 pll_reset : in std_logic; -- global PLL Reset
		 
		 conv_sync : in std_logic;
		 
		 adc_cnv   : out std_logic;-- convert signal out to ADC.
		 adc_clk   : out std_logic;-- data clock out to ADC.
		 raw_pll   : out std_logic; -- DIAGNOSTIC
		 raw_gate  : out std_logic; -- DIAGNOSTIC
		 pll_locked : out std_logic -- PLL Locked indication
		);
end AD7625_INTERFACE_v2;
	
architecture main of AD7625_INTERFACE_v2 is
------------------------------------------------------------
-- ALTERA COMPONENTS ---------------------------------------
-- The following component is the PLL clock multiplier that
-- produces the ADC data clock from the LogicClk.
-- It was created using the LPM Megafunction Wizard.
-- The following VHDL files were generated and must be included 
-- in the project.
--                              ADC_CLK_PLL.vhd
--
-- The Megafunction Wizard also generate *.cmp files for each
-- from which I cut and pasted the component declarations below.
----------------------------------------------------------------	
component ADC_CLK_PLL
	PORT
	(
		areset		: IN STD_LOGIC  := '0';
		inclk0		: IN STD_LOGIC  := '0';
		c0			: OUT STD_LOGIC ;  -- 2x inclk0
		locked		: OUT STD_LOGIC 
	);
end component;

	-- PLL declarations ----------------------------------------------
	signal pll_rdy    : std_logic;
	signal pll_c0     : std_logic;
	
	signal gate_clk   : std_logic; -- data clock gate set by state mach.	
	signal syncreg1   : std_logic; -- dff outputs that sync the gate_clk
	signal syncreg2   : std_logic;
	signal syncreg3   : std_logic;
	signal syncd_gate : std_logic; -- syncronized data clock gate

	-- Counter/Timer declarations --------------------------------------
	signal nclr_cnt    : std_logic;
	signal marker      : std_logic_vector(7 downto 0);
	signal got_cnt_flag: std_logic;

	-- Conversion Synchronization ------------------------------------
	signal cnv_sync_led  : std_logic_vector(1 downto 0);
	signal cnv_sync_strt : std_logic;
	signal cnv_sync_dly	 : std_logic_vector(4 downto 0);
	signal cnv_sync		 : std_logic;
	
	-- State Machine declarations ------------------------------------
	type STATE_TYPE is 
	(init, initial_pause, start, set_cnv0, wait0, set_cnv, wait1, 
		pll_off1, pll_off2, pll_strt, wait_sync, pause_sync);
	signal CS, NS  : STATE_TYPE;

--===begin===begin===begin===begin===begin===begin===begin===begin===
begin -- main --------------------------------------------------------
	ADC_CLK_PLL_inst : ADC_CLK_PLL 
	port map (
				areset	 => pll_reset,
				inclk0	 => clk,
				c0	 	 => pll_c0, -- 2x inclk0
				locked	 => pll_rdy
			 );		
	-------------------------------------------------------------------
	SEQCNT: counter  -- Timer
	generic map(WID8)
	port map ( clk		=> clk, 
			   clrn		=> nclr_cnt, 
			   ena		=> '1', 
			   load		=> '0', 
			   updn		=> '1',
			   ld_data  => (others => '0'),
			   flag_cnt => marker,
			   flag_out => got_cnt_flag,
			   cnt_out  => open
			 );

-- Conversion Sync Signals ---------------------------------------
process (clk, reset_n, conv_sync, cnv_sync_led)
begin
	if   (reset_n = '0') then cnv_sync_led <= (others => '0');
	elsif rising_edge(clk) then
		cnv_sync_led(0) <= conv_sync;
		cnv_sync_led(1) <= cnv_sync_led(0);
	end if;
	cnv_sync_strt <= cnv_sync_led(0) AND NOT cnv_sync_led(1);
end process;
	
process (clk, reset_n, cnv_sync_strt, cnv_sync_dly)
begin
	if   (reset_n = '0') then cnv_sync_dly <= (others => '0');
	elsif rising_edge(clk) then
		if   (cnv_sync_strt = '1') 	then cnv_sync_dly <= (others => '0');
		elsif(cnv_sync = '1')		then cnv_sync_dly <= cnv_sync_dly + 1;
		end if;
	end if;
	--------------------------------------------------------
	if (cnv_sync_dly = "10001") then cnv_sync <= '0'; -- change this to position strobe
	else							 cnv_sync <= '1';
	end if;
end process;

-- State Machine define initialization ---------------------------
	sm_reg: process (clk, reset_n)	
	begin
		if(reset_n = '0')        	then CS <= init;
		elsif(clk'event and clk='1')then CS <= NS;
		end if;
	end process; -- sm_reg
	
	sm_combinatorial: process (CS, got_cnt_flag, pll_rdy, cnv_sync)
	begin
		-- default signal values
		nclr_cnt  <= '1'; -- active = '0'
		marker    <= X"02";
		adc_cnv   <= '0';
		gate_clk  <= '0';	
		-------------------------
		case CS is
			when init  =>
				if(pll_rdy = '1') then 		NS <= initial_pause;
				else                   		NS <= init;
				end if;			    
				nclr_cnt <= '0';		    

			when initial_pause	 =>
				marker <= X"FE"; -- Initial pause allows other FPGA to get ready
				if(got_cnt_flag = '1') then NS <= start;
				else                      	NS <= initial_pause;
				end if;			    
			
			when start => 					NS <= set_cnv0;
				nclr_cnt <= '0';		    
			
			when set_cnv0  =>				NS <= wait0;
				adc_cnv  <= '1'; -- the first time CNV pulses 
				nclr_cnt <= '0'; --     the adc clock is not gated.	    
					
			when wait0  =>
				--Note that the count marker runs here, not in reset
				marker <= X"02"; -- Only 2 since gate sync process takes time
				if(got_cnt_flag = '1') then NS <= pll_off1;
				else                      	NS <= wait0;
				end if;			    
					
			when set_cnv   =>				
				gate_clk  <= '1';
				nclr_cnt <= '0';
				
				if (cnv_sync = '0')	then 
					adc_cnv  <= '1';   		NS <= wait1;
				else
					adc_cnv	 <= '0';		NS <= wait_sync;
				end if;
						    
			----------------------------------------------------
			when wait1  =>
				gate_clk  <= '1';
				--Note that the count marker
				marker <= X"02"; -- Only 2 since gate sync process takes time
				if(got_cnt_flag = '1') then NS <= pll_off1;
				else                      	NS <= wait1;
				end if;	
						    
			when pll_off1  =>				NS <= pll_off2;
				
			when pll_off2  =>				NS <= pll_strt;
				
			when pll_strt  => -- beginning of next serial transfer --
				gate_clk  <= '1';
				--Note that the count marker is set below
				marker <= X"08";
				if(got_cnt_flag = '1') then NS <= set_cnv;
				else                      	NS <= pll_strt;
				end if;			    
			-------------------------------------------------------
			when wait_sync =>
				gate_clk  <= '1';
				--Note that the count marker
				marker <= X"02"; -- Only 2 since gate sync process takes time
				if(got_cnt_flag = '1') then NS <= pause_sync;
				else                      	NS <= wait_sync;
				end if;	
			
			when pause_sync =>
				if (cnv_sync = '0')	then 	NS <= start;
				else						NS <= pause_sync;
				end if;
						    
			when others	  =>				NS <= init;		
		end case; 
				
	end process; -- sm_combinatorial
	
	-- Syncronized PLL Gate ----------------------------
--	process(pll_c0, reset_n, syncreg1, syncreg2, syncreg3, gate_clk)
--	begin
--		if(reset_n = '0')  then 
--			syncreg1 <= '0';
--			syncreg2 <= '0';
--			syncreg3 <= '0';
--		elsif falling_edge(pll_c0) then
--			syncreg1 <= gate_clk;
--			syncreg2 <= syncreg1;
--			syncreg3 <= syncreg2;
--		end if;
--		syncd_gate <= syncreg3;
--	end process;
	
	process(pll_c0, reset_n, syncreg1, syncreg2, syncreg3, gate_clk)
	begin 									-- using a synchronized reset in this version
		if falling_edge(pll_c0) then
			if(reset_n = '0')  then 
				syncreg1 <= '0';
				syncreg2 <= '0';
				syncreg3 <= '0';
			else 
				syncreg1 <= gate_clk;
				syncreg2 <= syncreg1;
				syncreg3 <= syncreg2;
			end if;
		end if;
		syncd_gate <= syncreg3;
	end process;
	
	pll_locked <= pll_rdy;
	adc_clk <= pll_c0 when syncd_gate = '1' else '0';
	
	raw_pll <= pll_c0;
	raw_gate <= syncd_gate;
	
end main;

