Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 
USE work.DDS_VME_CONSTANTS.all;
-------------------------------------------------------------------------------
-- MODULE DESCRIPTION, March 30, 2015
-- This module manages two buffer memories for the Bias Curve
-- and provides a process for playing out the Bias Curve when 
-- triggered.

-- One of the two buffers is where the data for the curve is 
-- read to be played out the DAC channel.  The other buffer is
-- the one that is available to be written with a new curve.
-- once a new curve is written the roles of the buffers are
-- swapped.

-- When a new curve is written into one of the buffers, there
-- must be atleast 5500 writes.  That is, enough 6us updates
-- to cover 33ms.  Any less than this will not cover the whole
-- active part of the Booster cycle.  It is expected that the
-- curve will be written to cover the full 66ms interval, but
-- coverage beyond 35 ms has not been supported before this
-- version.

-- A/B Management: At reset and bootup initialization of memory
-- Memory A is written.  When the logic, here, see that 5500 
-- points have been written the read_A_nB bit will be set to 
-- read from Memory A.  The Write_A_nB bit is set to be the
-- inverse of the read_A_nB signal at the leading edge of 
-- Run_BiasCrv.

-- Version 2:  The Bias Curve Gate was added.  The gate allows
-- Run_BiasCrv to start the state machine.

entity Bias_Curve_Player_v2 is 
port( 
		clk						: in std_logic;
		nreset					: in std_logic;
		Run_BiasCrv				: in std_logic;
		BiasCrv_Gate			: in std_logic;
		BiasCrv_From_Mem_A	: in std_logic_vector(15 downto 0);
		BiasCrv_From_Mem_B	: in std_logic_vector(15 downto 0);
		UPPER_WR_BC				: in std_logic;  -- we will count the writes
		
		Write_A_nB	 			: out std_logic;
		BiasCrv_Addr_to_Mem 	: out std_logic_vector(23 downto 0); -- address goes to A and B
		Crv_Data_to_DAC_OB	: out std_logic_vector(15 downto 0);
		DAC_Strobe_Out			: out std_logic
		
	);
end Bias_Curve_Player_v2;
	
architecture main of Bias_Curve_Player_v2 is	

-- Internal Signals ------------------------------------------
	signal go				: std_logic;
	signal go_0				: std_logic;
	signal curve_done		: std_logic;
	signal read_A_nB		: std_logic;

-- Pause Timer -----------------------------------------------
	signal pause_tmr 		: std_logic_vector(4 downto 0);
	signal inc_pause_tmr	: std_logic;
	
-- Update Timer ----------------------------------------------
	signal update_tmr		: std_logic_vector(8 downto 0);
	signal update_tmr_ena: std_logic;
	signal update			: std_logic;
	
-- Count Memory Writes ---------------------------------------
	signal bc_write		: std_logic;
	signal wr_count		: std_logic_vector(12 downto 0);
	signal wr_count_ena	: std_logic;
	signal wr_count_met	: std_logic;
	
	signal last_bc_point	: std_logic;
	
-- Bias Curve Pointer ----------------------------------------
		signal inc_biascrv_ptr	: std_logic;
		signal rst_biascrv_ptr	: std_logic;
		signal biascrv_ptr		: std_logic_vector(23 downto 0); 
		
-- Bias Curve Register ---------------------------------------
	signal bias_curve_reg		: std_logic_vector(15 downto 0);
	signal get_bias_curve		: std_logic;

-- SETUP_STATE_MACHINE ---------------------------------------
type STATE_TYPE is
(init, init_dac, idle, get_curve_point, strobe_dac, 
										increment_pointer, wait_4us, done);
signal CS, NS : STATE_TYPE;

--=====================================================================
--				B  E  G  I  N 
--=====================================================================
begin

-- LED SYNCHRONIZATION --------------------------------  
SYNC_RUN: led -- syncronize the signal to start playing the Bias Curve.
port map(sig_in  => Run_BiasCrv, clk => clk, clrn => '1', pls_out => go_0); 

go <= go_0 AND BiasCrv_Gate;

SYNC_WR: led
port map(sig_in  => UPPER_WR_BC, clk => clk, clrn => '1', pls_out => bc_write); 

--mmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmm
-----------------------------------------------------------------------------
PAUSE_TIMER: -- Timer used to generate wait states for memory reads
process (clk, inc_pause_tmr)
begin
	if rising_edge(clk) then
		if   (inc_pause_tmr = '0') then pause_tmr <= "00000";
		elsif(inc_pause_tmr = '1') then pause_tmr <= pause_tmr + 1;
		end if;
	end if;
end process;

------------------------------------------------------------------------
-- 4 us UPDATE_TIMER ---------------------------------------------------
UPDATE_TIMER: -- Timer used to signal DAC Updates 
-- X"C8" = 200, 200 * 20ns_clk = 4us
process (clk, update_tmr, update_tmr_ena)
begin
	if rising_edge(clk) then
		if   (update_tmr_ena = '0') then update_tmr <= (others => '0');
													update <= '0';
		elsif(update_tmr = X"C8")   then update_tmr <= (others => '0');
													update <= '1';
		else 										update_tmr <= update_tmr + 1;
													update <= '0';
		end if;
	end if;
end process;

-----------------------------------------------------------------------
-- ADDRESS_POINTER ----------------------------------------------------
BIAS_CURVE_ADDRESS_PTR:
process(clk, nreset, inc_biascrv_ptr, rst_biascrv_ptr, biascrv_ptr)
begin
	if  (nreset = '0') then 				biascrv_ptr <= BIAS_CURVE_ADDR;
	elsif rising_edge(clk) then
		if   (rst_biascrv_ptr = '1') then biascrv_ptr <= BIAS_CURVE_ADDR;
		elsif(inc_biascrv_ptr = '1') then biascrv_ptr <= biascrv_ptr + 2;
		end if;
	end if;
end process;

process(clk, nreset, biascrv_ptr)
begin
	if  (nreset = '0') then 				last_bc_point <= '0';
	elsif rising_edge(clk) then
		if   (biascrv_ptr = BIAS_CURVE_LAST) then 
													last_bc_point <= '1';
		else										last_bc_point <= '0';
		end if;
	end if;
end process;

	BiasCrv_Addr_to_Mem <= biascrv_ptr(23 downto 0);

-----------------------------------------------------------------------
-- MEMORY_WRITE_COUNTER -------------------------------------------------
-- This will count the memory writes to the Bias Curve Buffer specified
-- by the "Write_A_nB" signal (inverse of the "read_A_nB" signal).  The
-- Curve Interpolation Process in the UPPER FPGA is what writes this 
-- memory during the update period at the end of the active Booster cycle.
-- That is, approx. 38ms after a TClk trigger (or BDot trigger) signaling 
-- the start of a cycle.  It is expected that the Curve Interpolation and
-- the writing of the Bias Curve buffer is completed before the start of 
-- the next cycle.  Note however, that the Curve Interpolation process
-- will only process one curve per cycle so that the updating of then
-- new curve sent from ACNET may not occur during the very next cycle.
MEMORY_WR_COUNTER:
process(clk, nreset, bc_write, wr_count_ena, wr_count)
begin
	if  (nreset = '0') 						 then wr_count <= (others => '0');
	elsif rising_edge(clk) then
		if   (wr_count_ena = '0')         then wr_count <= (others => '0');
		elsif(wr_count = "1010101111100") then wr_count <= wr_count;
		elsif(bc_write     = '1')         then wr_count <= wr_count + 1;
		end if;
	end if;
end process;

-- Since the Curve Interpolation is expected to update the Bias Curve
-- buffer (16k points) within one Booster cycle.  The 'wr_count' is
-- cleared at the end of each Booster cycle.

process(clk, nreset, wr_count)
begin
	if  (nreset = '0') then 		wr_count_met <= '0';
	elsif rising_edge(clk) then
		if   (wr_count = "1010101111100") then -- count = 5500 writes
											wr_count_met <= '1';
		else								wr_count_met <= '0';
		end if;
	end if;
end process;

-----------------------------------------------------------------------
-- READ_A_nB_REGISTER -------------------------------------------------
-- The toggling of this register happens only when "wr_count_met" overlaps
-- with the pulsing of "curve_done".  "wr_count" is cleared immediately
-- after this one clock wide pulse of "wr_count_met AND curve_done".
READ_A_nB_REGISTER:
process(clk, nreset, read_A_nB, wr_count_met, curve_done)
begin
	if  (nreset = '0') 				then read_A_nB <= '1';
	elsif rising_edge(clk) then
		if   (wr_count_met = '1') 	then 
			if(curve_done = '1')   	then read_A_nB <= not read_A_nB;
			else								  read_A_nB <=     read_A_nB;
			end if;
		end if;
	end if;
end process;

	Write_A_nB <= not read_A_nB; --<<<<<<<<< OUTPUT <<<<<<<<<<<<<<<<<<<
	
----------------------------------------------------------------------
-- BIAS CURVE REGISTER -----------------------------------------------
-- The next curve point from memory is latched here.

BIAS_CURVE_REGISTER:
process(clk, nreset, get_bias_curve, read_A_nB, BiasCrv_From_Mem_A,
			BiasCrv_From_Mem_B)
begin
	if  (nreset = '0') then 			   	bias_curve_reg <= (others => '0');
	elsif rising_edge(clk) then
		if   (get_bias_curve = '1') then 
			if (read_A_nB = '1') 	 then		bias_curve_reg <= BiasCrv_From_Mem_A;
			else										bias_curve_reg <= BiasCrv_From_Mem_B;
			end if;
		end if;
	end if;
end process;

	Crv_Data_to_DAC_OB <= not bias_curve_reg(15) & bias_curve_reg(14 downto 0); --<<<<<<<<< OUTPUT <<<<<<<<<<<<<<<<<<<

--{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{
--                S T A T E   M A C H I N E
--}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}
SM_REG: process(clk, nreset)
begin
	if(nreset='0') then CS <= init;
	elsif(rising_edge(clk))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process
(CS, go, last_bc_point, pause_tmr, update)
begin
	-- defaults ----------------------
	get_bias_curve		<= '0';
	
	curve_done			<= '0';
	
	wr_count_ena		<= '1'; -- disabled only in a few states (eg idle)
	
	rst_biascrv_ptr	<= '0';
	inc_biascrv_ptr	<= '0';
	
	inc_pause_tmr		<= '0';
	update_tmr_ena 	<= '0';
	
	DAC_Strobe_Out		<= '0';
	-----------------------------------
	case CS is
		when init =>  -- get first point and setup dac output
			wr_count_ena		<= '0';
			rst_biascrv_ptr	<= '1';
			get_bias_curve		<= '1'; -- data latches into register when set to '0' next
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "00011") then		NS <= init_dac;
			else										NS <= init;
			end if;				
				
		when init_dac =>  -- setup dac output	
			wr_count_ena		<= '0';
			DAC_Strobe_Out		<= '1';
														 NS <= idle;
														 
		----------------------------------------------------------------------
		when idle =>	 -- we wait here after the first time through       
			rst_biascrv_ptr	<= '1';
			wr_count_ena		<= '1'; -- enabled, paying attention to buffer writes now.
			
			if (go = '1') then			NS <= get_curve_point;
			else								NS <= idle;
			end if;	
			
		----------------------------------------------------------------------
		when get_curve_point =>  -- get first point and setup dac output
			update_tmr_ena 	<= '1';
			get_bias_curve		<= '1'; -- data latches into register when set to '0' next
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "00011") then		NS <= strobe_dac;
			else										NS <= get_curve_point;
			end if;				
		
		when strobe_dac => 						NS <= increment_pointer;
			update_tmr_ena 	<= '1';
			DAC_Strobe_Out		<= '1';
			
		when increment_pointer =>
			update_tmr_ena 	<= '1';
			if   (last_bc_point = '0') then 	NS <= wait_4us; -- not last point
					inc_biascrv_ptr	<= '1';
			else									 	NS <= done; -- last point
					inc_biascrv_ptr	<= '0';
			end if;
		
		when wait_4us =>
			update_tmr_ena 	<= '1';
			
			if   (update = '1') then 			NS <= get_curve_point;
			else										NS <= wait_4us;
			end if;
			
		-----------------------------------------------
		when done =>								NS <= idle;
				curve_done		<= '1';  -- toggle read_A_nB if new curve was written
				wr_count_ena	<= '0';  -- clears the 'write count'

		when others => 							NS <= idle;
	end case;
end process;

end;