Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 

------------------------------------------------------
-- MODULE DESCRIPTION
-- This is mainly logic and timing for writing the serial 
-- control interface of the AD9910.

-- This module has several parts
--
--	1.  There is a state machine with 4 simple states; 
--		idle, first_bit, send, update.
--	2.	There is a 72 bit shift register that is parallel loaded when 
--		the state switches from idle to first_bit, and shifts data out 
--		on each clock when in the send state.
--	3.  There is a shift register counter to signal when the desired number of
--		bits (BitCount) has been sent.
--	4.  Finally, there are strings of flip-flops to syncronize the final
-- 		data update signals.  SYNC_IO_UPDATE set of dffr's syncronizes the
--		IO_Update signal to the DDS Sync_Clock.  SYNC_TXDONE set of dffr's
--		takes the final output of SYNC_IO_UPDATE and re-syncronizes this edge
--		to the logic clock, SR_Clk, to produce a transmission complete signal,
--		tx_done.  tx_done is also sent out as the IO_Reset signal when the
--		IO_Reset signal is desired.
--
------------------------------------------------------

entity AD9910_Serial_Intf72_v3 is 
	port( 
		SR_Clk		: in std_logic; -- clock that drives the Shift Reg
		nreset		: in std_logic;

		Sync_Clk_1  : in std_logic;
		Sync_Clk_2  : in std_logic;
		Sync_Clk_3  : in std_logic;
		Sync_Clk_4  : in std_logic;
		
		TxStrb		: in std_logic; -- strobe to latch data to transmit
		TxData_1	: in std_logic_vector(71 downto 0); 
		TxData_2	: in std_logic_vector(71 downto 0); 
		TxData_3	: in std_logic_vector(71 downto 0); 
		TxData_4	: in std_logic_vector(71 downto 0); 
		
		BitCount	: in std_logic_vector(6 downto 0);
		Ena_IOReset	: in std_logic; -- used to send IO_Reset (optional)
		
		DDS_CS_1	: in std_logic; -- module chip select
		DDS_CS_2	: in std_logic; -- module chip select
		DDS_CS_3	: in std_logic; -- module chip select
		DDS_CS_4	: in std_logic; -- module chip select
		
		DDS_SOUT_1	: out std_logic; -- serial data to DDS
		DDS_SCLK_1	: out std_logic; -- serial clock
		DDS_nCS_1	: out std_logic; -- chip select
		IO_Update_1	: out std_logic;
		IO_Reset_1  : out std_logic; -- IO_Reset terminates a transmission
		
		DDS_SOUT_2	: out std_logic; -- serial data to DDS
		DDS_SCLK_2	: out std_logic; -- serial clock
		DDS_nCS_2	: out std_logic; -- chip select
		IO_Update_2	: out std_logic;
		IO_Reset_2  : out std_logic; -- IO_Reset terminates a transmission
		
		DDS_SOUT_3	: out std_logic; -- serial data to DDS
		DDS_SCLK_3	: out std_logic; -- serial clock
		DDS_nCS_3	: out std_logic; -- chip select
		IO_Update_3	: out std_logic;
		IO_Reset_3  : out std_logic; -- IO_Reset terminates a transmission
		
		DDS_SOUT_4	: out std_logic; -- serial data to DDS
		DDS_SCLK_4	: out std_logic; -- serial clock
		DDS_nCS_4	: out std_logic; -- chip select
		IO_Update_4	: out std_logic;
		IO_Reset_4  : out std_logic; -- IO_Reset terminates a transmission
		
	tx_out : out std_logic;	
		Transfer_Done	: out std_logic  -- Strobes when SR is done
		);
	end AD9910_Serial_Intf72_v3;
	
---------------------------------------------------------
architecture main of AD9910_Serial_Intf72_v3 is
-- SYNC_TX Signals ---------------------------------------
	signal sync_tx_d : std_logic_vector(2 downto 0);
	signal sync_tx_q : std_logic_vector(2 downto 0);
	signal tx		 : std_logic; --run transmit SR
		
-- SYNC_IO_UPDATE Signals --------------------------------
	signal IOU1_d		 : std_logic_vector(4 downto 0);
	signal IOU1_q		 : std_logic_vector(4 downto 0);
	signal IOU2_d		 : std_logic_vector(4 downto 0);
	signal IOU2_q		 : std_logic_vector(4 downto 0);
	signal IOU3_d		 : std_logic_vector(4 downto 0);
	signal IOU3_q		 : std_logic_vector(4 downto 0);
	signal IOU4_d		 : std_logic_vector(4 downto 0);
	signal IOU4_q		 : std_logic_vector(4 downto 0);
	
	signal IOU_enable 	 : std_logic;
	signal IOU_clrn		 : std_logic;
	signal IOUx_clrn		 : std_logic;
	signal IOU_syncd     : std_logic;
	signal IOU1, IOU2, IOU3, IOU4 : std_logic;
	signal IOU_Falling_Edge : std_logic;

	signal TXD_d		 : std_logic_vector(2 downto 0);
	signal TXD_q		 : std_logic_vector(2 downto 0);

	signal tx_done		 : std_logic;
		
-- TRANSMIT_SR Signals -----------------------------------
	signal xmt_sr1_d	: std_logic_vector(71 downto 0);
	signal xmt_sr1_q	: std_logic_vector(71 downto 0);
	signal xmt_sr2_d	: std_logic_vector(71 downto 0);
	signal xmt_sr2_q	: std_logic_vector(71 downto 0);
	signal xmt_sr3_d	: std_logic_vector(71 downto 0);
	signal xmt_sr3_q	: std_logic_vector(71 downto 0);
	signal xmt_sr4_d	: std_logic_vector(71 downto 0);
	signal xmt_sr4_q	: std_logic_vector(71 downto 0);
	
	signal xmt_sr_ena	: std_logic;
	
	signal nsclk_ena	: std_logic; --active low SCLK gate

-- SR_COUNT Signals --------------------------------------
	signal sr_cnt_nrst	: std_logic;
	signal sr_cnt_ena	: std_logic;
	signal sr_done		: std_logic;
	signal sr_cnt_flag  : std_logic_vector(6 downto 0);
	
-- SR_STATE_MACHINE --------------------------------------
	type STATE_TYPE is
		(idle, first_bit, send, update);
	signal CS, NS : STATE_TYPE;	
	
--========================================================
--                  B E G I N
--========================================================
begin

--======================================================================
-- INPUT STROBE SYNCRONIZATION
--======================================================================  
SYNC_TX: dffr -- Syncronize tx
generic map(3)
port map(	d    => sync_tx_d(2 downto 0),
			clk  => SR_Clk,
			clrn => '1', -- note active low reset
			ena  => '1',
			prn  => '1',
			q    => sync_tx_q(2 downto 0)
		); 

	sync_tx_d(0)  <= TxStrb;
	sync_tx_d(2 downto 1)  <= sync_tx_q(1 downto 0);
	tx <= sync_tx_q(0) AND (NOT sync_tx_q(2));
--> CONSIDER WHETHER WE REALLY NEED THIS PULSE TWO CLOCKS WIDE (?).
 
--=======================================================================
-- STATE MACHINE  
--=======================================================================
SM_REG: 
process(SR_Clk, nreset)
begin
	if(nreset='0')				then CS <= idle;
	elsif(rising_edge(SR_Clk))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process(CS, tx, sr_done, tx_done)
begin
	-- Default Values ----------------------------
			sr_cnt_nrst <= '1';
			sr_cnt_ena  <= '0';
			
	case CS is
		when idle => 
			-- In this state we are waiting to see the tx strobe
			-- telling us to transmit the data.  While waiting
			-- the shift registers are ready for a parallel load
			-- of the data, and the tx pulse will latch this data
			-- when it comes along.
			sr_cnt_nrst <= '0'; -- sr count is in reset
			sr_cnt_ena  <= '0'; -- sr count disabled
			
			if   (tx = '1')   then 	NS <= first_bit;
			else                   	NS <= idle;
			end if;
			
		when first_bit => 			NS <= send;
			-- In this state we clock out the first bit WITHOUT
			-- advancing the register on the falling edge of the SR clock.
			sr_cnt_ena  <= '1';
		
		when send => 
			-- In this state bits shift out of registers until the
			-- shift register counter says we are done (sr_done).
			sr_cnt_ena  <= '1';
			
			if (sr_done = '1') then NS <= update;
			else					NS <= send;
			end if;
			
		when update => 
			-- In this state we are waiting until the IO_Update
			-- pulses have all been sent out resulting in tx_done.
			if (tx_done = '1') then NS <= idle;
			else					NS <= update;
			end if;
			
		when others =>              NS <= idle;
	end case;
end process;

--==========================================================================
-- TRANSMIT SHIFT REGISTERS
--   This section defines the four shift registers and how they respond to
--   the state machine state transitions.
--==========================================================================  
TRANSMIT_SR1: dffr -- the transmit shift register
generic map(72)
port map(	d    => xmt_sr1_d(71 downto 0),
			clk  => not SR_Clk,
			clrn => nreset,
			ena  => xmt_sr_ena,
			prn  => '1',
			q    => xmt_sr1_q(71 downto 0) ); 
			
--**--   --**--   --**--   --**--   --**--   --**--   --**--    
TRANSMIT_SR2: dffr -- the transmit shift register
generic map(72)
port map(	d    => xmt_sr2_d(71 downto 0),
			clk  => not SR_Clk,
			clrn => nreset,
			ena  => xmt_sr_ena,
			prn  => '1',
			q    => xmt_sr2_q(71 downto 0) ); 
			
--**--   --**--   --**--   --**--   --**--   --**--   --**--    
TRANSMIT_SR3: dffr -- the transmit shift register
generic map(72)
port map(	d    => xmt_sr3_d(71 downto 0),
			clk  => not SR_Clk,
			clrn => nreset,
			ena  => xmt_sr_ena,
			prn  => '1',
			q    => xmt_sr3_q(71 downto 0) ); 
			
--**--   --**--   --**--   --**--   --**--   --**--   --**--    
TRANSMIT_SR4: dffr -- the transmit shift register
generic map(72)
port map(	d    => xmt_sr4_d(71 downto 0),
			clk  => not SR_Clk,
			clrn => nreset,
			ena  => xmt_sr_ena,
			prn  => '1',
			q    => xmt_sr4_q(71 downto 0) ); 
			
-------------------------------------------------------------------------   
STATE_MACHINE_CONTROL_OF_TRANSMIT_SRs: 
-------------------------------------------------------------------------
process(CS, tx, TxData_1, TxData_2, TxData_3, TxData_4, 
			xmt_sr1_q, xmt_sr2_q, xmt_sr3_q, xmt_sr4_q,
			DDS_CS_1, DDS_CS_2, DDS_CS_3, DDS_CS_4)
begin
	-- Default Values ---------------------------------------------------
	-- bits were reveresed in previous modules if desiring LSB first
	xmt_sr1_d(71 downto 1) <= xmt_sr1_q(70 downto 0); xmt_sr1_d(0)<= '0';
	DDS_SOUT_1   <= xmt_sr1_q(71);
	DDS_nCS_1  <= NOT DDS_CS_1;
	
	xmt_sr2_d(71 downto 1) <= xmt_sr2_q(70 downto 0); xmt_sr2_d(0)<= '0';
	DDS_SOUT_2   <= xmt_sr2_q(71);
	DDS_nCS_2  <= NOT DDS_CS_2;
	
	xmt_sr3_d(71 downto 1) <= xmt_sr3_q(70 downto 0); xmt_sr3_d(0)<= '0';
	DDS_SOUT_3   <= xmt_sr3_q(71);
	DDS_nCS_3  <= NOT DDS_CS_3;
	
	xmt_sr4_d(71 downto 1) <= xmt_sr4_q(70 downto 0); xmt_sr4_d(0)<= '0';
	DDS_SOUT_4   <= xmt_sr4_q(71);
	DDS_nCS_4  <= NOT DDS_CS_4;		
	---------------------------------------------------------------------

	case CS is
		when idle => 
			xmt_sr_ena <= tx;     --tx is the syncronized transmit strobe 
								  -- It latches parallel data into SR
			nsclk_ena  <= '1'; 	  --Clock is Masked 
			
			xmt_sr1_d   <= TxData_1; --The shift register is loaded with TxData(71..0)
			DDS_SOUT_1  <= '0';	  	 --Serial Data Output to DDS	
			DDS_nCS_1   <= '1'; 	  --Chip Select is Not Active
			
			xmt_sr2_d   <= TxData_2; --The shift register is loaded with TxData(71..0)
			DDS_SOUT_2  <= '0';	  	 --Serial Data Output to DDS	
			DDS_nCS_2   <= '1'; 	  --Chip Select is Not Active
			
			xmt_sr3_d   <= TxData_3; --The shift register is loaded with TxData(71..0)
			DDS_SOUT_3  <= '0';	  	 --Serial Data Output to DDS	
			DDS_nCS_3   <= '1'; 	  --Chip Select is Not Active
			
			xmt_sr4_d   <= TxData_4; --The shift register is loaded with TxData(71..0)
			DDS_SOUT_4  <= '0';	  	 --Serial Data Output to DDS	
			DDS_nCS_4   <= '1'; 	  --Chip Select is Not Active
			
		when first_bit => 
			xmt_sr_ena <= tx;	--Shift Register Clock Disabled  ---!@#$!@#$!@#$!@#$!@#$   ??????
			nsclk_ena  <= '0'; --DDS_SCLK <= clk OR nsclk_ena; -- Clock Unmasked 
			
		when send => 
			xmt_sr_ena <= '1';	--Shift Register Clock Enabled
			nsclk_ena  <= '0'; 	--DDS_SCLK <= clk OR nsclk_ena; -- Clock Unmasked 
			
		when update => 
			xmt_sr_ena <= '0';	--Shift Register Clock Disabled
			nsclk_ena  <= '1'; 	--Clock is Masked 
	end case;
end process;

		DDS_SCLK_1 <= SR_Clk OR nsclk_ena; -- Clock Stall High method
		DDS_SCLK_2 <= SR_Clk OR nsclk_ena; -- Clock Stall High method
		DDS_SCLK_3 <= SR_Clk OR nsclk_ena; -- Clock Stall High method
		DDS_SCLK_4 <= SR_Clk OR nsclk_ena; -- Clock Stall High method
			
--====================================================================
-- SHIFT REGISTER BIT COUNTER
--   Signals with sr_done when the desired number of bits (BitCount)
--   have been clocked out.
--====================================================================
SR_COUNT: counter -- Counter to control the Transmit SR's
generic map(7)
port map(	clk => SR_Clk,
			clrn => sr_cnt_nrst, -- SM controlled
			ena  => sr_cnt_ena,  -- SM controlled
			load => '0',
			updn => '1',
			ld_data  => (others => '0'),
			flag_cnt => BitCount,
			flag_out => sr_done,
			cnt_out  => OPEN        );

--============================================================
-- GENERATE SYNCRONIZED IO_UPDATE PULSES 
--   The only input to this section is the signal sr_done
--============================================================
--**-- --**--   --**--   --**--   --**--   --**--   --**--   
LATCH_SRDONE: vdffe -- latch sr_done until tx_done is signaled
	port map( d    => '1',
			  clk  => SR_Clk,
			  clrn => IOU_clrn, -- note active low reset
			  prn  => '1',
			  ena  => sr_done, -- this is flag_out of the SR_COUNT
			  q    => IOU_enable );
			  
	IOU_clrn <= NOT TXD_q(2) AND nreset; -- See TXD registers below
	
--**-- --**--   --**--   --**--   --**--   --**--   --**--   
SYNC_IO_CLEAR: vdffe -- Syncronize to SYNC_CLK
port map(	d    => IOU_clrn,
			clk  => Sync_Clk_1,
			clrn => nreset, -- note active low reset
			ena  => '1',
			prn  => '1',
			q    => IOUx_clrn ); 	

------------------------------------------------------------------
	-- IOU_syncd ensures that all IO_Update timing registers start 
	-- together.  Having the IOU_enable in common does not ensure this,
	-- since the IOU_enable is syncronous to SR_Clk and not Sync_Clk_x.
	-- Note, still, IOU_syncd does not ensure the IO_Update timing is
	-- syncronized before the various Sync_Clk's are in sync themselves.
	
	IOU_syncd <= IOU1_q(0) AND IOU2_q(0) AND IOU3_q(0) AND IOU4_q(0); 
	IOU_Falling_Edge <= IOU1 OR IOU2 OR IOU3 OR IOU4;
	IO_Update_1 <= IOU_Falling_Edge;
	IO_Update_2 <= IOU_Falling_Edge;
	IO_Update_3 <= IOU_Falling_Edge;
	IO_Update_4 <= IOU_Falling_Edge;
			  
--**-- --**--   --**--   --**--   --**--   --**--   --**--   
SYNC_IO_UPDATE_1: dffr -- Syncronize to SYNC_CLK
generic map(5)
port map(	d    => IOU1_d(4 downto 0),
			clk  => Sync_Clk_1,
			clrn => IOUx_clrn, -- note active low reset
			ena  => IOU_enable,
			prn  => '1',
			q    => IOU1_q(4 downto 0) ); 

	IOU1_d(0)  <= '1';
	IOU1_d(1)  <= IOU_syncd;
	IOU1_d(4 downto 2)  <= IOU1_q(3 downto 1);
	IOU1 <= IOU1_q(1) AND (NOT IOU1_q(4));
		
--**-- --**--   --**--   --**--   --**--   --**--   --**--   
SYNC_IO_UPDATE_2: dffr -- Syncronize to SYNC_CLK
generic map(5)
port map(	d    => IOU2_d(4 downto 0),
			clk  => Sync_Clk_2,
			clrn => IOUx_clrn, -- note active low reset
			ena  => IOU_enable,
			prn  => '1',
			q    => IOU2_q(4 downto 0) ); 

	IOU2_d(0)  <= '1';
	IOU2_d(1)  <= IOU_syncd;
	IOU2_d(4 downto 2)  <= IOU2_q(3 downto 1);
	IOU2 <= IOU2_q(1) AND (NOT IOU2_q(4));
		
--**-- --**--   --**--   --**--   --**--   --**--   --**--   
SYNC_IO_UPDATE_3: dffr -- Syncronize to SYNC_CLK
generic map(5)
port map(	d    => IOU3_d(4 downto 0),
			clk  => Sync_Clk_3,
			clrn => IOUx_clrn, -- note active low reset
			ena  => IOU_enable,
			prn  => '1',
			q    => IOU3_q(4 downto 0) ); 

	IOU3_d(0)  <= '1';
	IOU3_d(1)  <= IOU_syncd;
	IOU3_d(4 downto 2)  <= IOU3_q(3 downto 1);
	IOU3 <= IOU3_q(1) AND (NOT IOU3_q(4));
	
--**-- --**--   --**--   --**--   --**--   --**--   --**--   
SYNC_IO_UPDATE_4: dffr -- Syncronize to SYNC_CLK
generic map(5)
port map(	d    => IOU4_d(4 downto 0),
			clk  => Sync_Clk_4,
			clrn => IOUx_clrn, -- note active low reset
			ena  => IOU_enable,
			prn  => '1',
			q    => IOU4_q(4 downto 0) ); 

	IOU4_d(0)  <= '1';
	IOU4_d(1)  <= IOU_syncd;
	IOU4_d(4 downto 2)  <= IOU4_q(3 downto 1);
	IOU4 <= IOU4_q(1) AND (NOT IOU4_q(4));
	
--**-- --**--   --**--   --**--   --**--   --**--   --**--   
SYNC_TXDONE: dffr -- Syncronize tx_done back to SR_Clk
generic map(3)
port map(	d    => TXD_d(2 downto 0),
			clk  => SR_Clk,
			clrn => nreset,
			ena  => '1',
			prn  => '1',
			q    => TXD_q(2 downto 0) ); 
			
	TXD_d(0) <= IOU1_q(4) AND IOU2_q(4) AND IOU3_q(4) AND IOU4_q(4); 
	TXD_d(2 downto 1) <= TXD_q(1 downto 0);
	
	tx_done    <= TXD_q(1) AND (NOT TXD_q(2));
	
	IO_Reset_1 <= TXD_q(1) AND (NOT TXD_q(2)) AND Ena_IOReset;
	IO_Reset_2 <= TXD_q(1) AND (NOT TXD_q(2)) AND Ena_IOReset;
	IO_Reset_3 <= TXD_q(1) AND (NOT TXD_q(2)) AND Ena_IOReset;
	IO_Reset_4 <= TXD_q(1) AND (NOT TXD_q(2)) AND Ena_IOReset;
							
	Transfer_Done 	<= TXD_q(1); -- CONSIDER THIS
				-- Note this stays active for 2 to 3 SR_CLK's.

tx_out <= tx; -- temporary diagnostic

end main;		
		
			