Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 
USE work.DDS_VME_CONSTANTS.all;
------------------------------------------------------
------------------------------------------------------
-- MODULE DESCRIPTION May 24, 2010
--
-- This module implements and manages bootups and backups
-- with the Flash memory.  When the module is first powered
-- (bootup) the shared memory will be loaded from the values
-- that are stored in the Flash memory.  Once the shared memory
-- is loaded, a global update process runs that takes the 
-- shared memory parameters and curves and initializes the 
-- run mode process registers and memory.
--
-- This module will also copy the shared memory to a Flash
-- memory block when triggered by the Backup_Trigger signal.
-- 
-- Two separate Flash memory blocks are used to store and 
-- restore the shared memory.  The object of this is to 
-- reduce the probability that the stored block of values
-- in the Flash should be corrupted due to loss of crate power
-- during a backup.  Storage of the shared memory is made 
-- alternately to the Flash memory block at byte addresses 0x0A0000 and at
-- 0x0C0000. The 16 bit addresses used to address the Flash are
-- 0x050000 and 0x060000.  The last two 32 bit words in 
-- the shared memory,and in the first and second Flash memory 
-- blocks, holds an "increment".  Each time the shared memory is 
-- stored to one of the Flash memory blocks, that blocks "increment" 
-- word is incremented.  FLASH BLOCK 1 IS USED FIRST WHEN THE MODULE
-- IS INITIALIZED.	So, when the first Flash block's increment
-- is greater than the second Flash block's increment we will
-- boot from the first, and the next backup will be made to the
-- second.  When a successful backup is made to the second
-- Flash block, its increment word is incremented so its 
-- increment word will now be greater than the firsts.
--

entity Bootup_Backup_Process_v3 is 
	port( 
		clk					: in std_logic;
		nreset				: in std_logic;
		UPD_Data_From_Mem 	: in std_logic_vector(31 downto 0);
		Bootup_Trigger		: in std_logic;
		Backup_Trigger		: in std_logic;		
		
		Flash_Operation_Done: in std_logic; -- Strobes when Unlock or Erase is done
		Flash_Transfer_Done	: in std_logic; -- Strobes when FLASH is done
		-----------------------------------------------------------------
		UPD_Data_to_Mem		: out std_logic_vector(31 downto 0);
		UPD_Addr_to_Mem		: out std_logic_vector(23 downto 0);
		UPD_WR_Pulse		: out std_logic;
		UPD_Bus_Ena			: out std_logic;

		RAM_Addr_In			: out std_logic_vector(23 downto 0);		
		FLASH_Addr			: out std_logic_vector(24 downto 1);		
		NWords				: out std_logic_vector(15 downto 0);
		Unlck_Block			: out std_logic;  -- trigger to unlock flash block
		Erase_Block			: out std_logic;  -- trigger to erase flash block
		Write_Flash			: out std_logic;  -- trigger for Write process
		Read_Flash			: out std_logic;  -- trigger for Read process
		
		Backup_Setup_Done	: out std_logic; -- signals bus available until Bus Request/Bus Permit
		Process_Done		: out std_logic;
		test_out			: out std_logic;
		last_store_bit		: out std_logic;
		last_restore_bit	: out std_logic;
		state					: out std_logic_vector(3 downto 0)
		);
	end Bootup_Backup_Process_v3;
	
---------------------------------------------------------
architecture main of Bootup_Backup_Process_v3 is
-- SYNC_DO_TEST Signals --------------------------------
	signal do_bootup	: std_logic;
	signal do_backup	: std_logic;
	
	signal cmd_bootup	: std_logic;
	signal cmd_backup	: std_logic;

	signal clear_bootup	: std_logic;
	signal clear_backup	: std_logic;

-- Data Register 1 --------------------------------------------------
	signal get_incr_1	: std_logic;
	signal clrn_data1	: std_logic;
	signal data_reg1_d	: std_logic_vector(31 downto 0);
	signal data_reg1_q	: std_logic_vector(31 downto 0);
	
	signal incr_block1  : std_logic_vector(31 downto 0);

-- Data Register 2 --------------------------------------------------
	signal get_incr_2	: std_logic;
	signal clrn_data2	: std_logic;
	signal data_reg2_d	: std_logic_vector(31 downto 0);
	signal data_reg2_q	: std_logic_vector(31 downto 0);

	signal incr_block2  : std_logic_vector(31 downto 0);

-- Data Register 3 --------------------------------------------------
	signal get_sm_incr	: std_logic;
	signal set_sm_1		: std_logic;
	signal set_sm_2		: std_logic;
	signal inc_sm_incr	: std_logic;

	signal sm_incr		: std_logic_vector(31 downto 0);
	
-- Pause Timer ------------------------------------
	signal pause_tmr 	: std_logic_vector(2 downto 0);
	signal inc_pause_tmr: std_logic;

-- FLASH Address Variable -----------------------------------------------
	signal FLASH_Addr_In : std_logic_vector(23 downto 0);
	signal converted_Addr: std_logic_vector(24 downto 1);

-- Last Store and Restore Bit Register ----------------------------------
	signal get_last_store	: std_logic;
	signal last_store	  	: std_logic;
	    
	signal get_last_restore	: std_logic;
	signal last_restore	  	: std_logic;
	    
-- BOOTUP BACKUP STATE_MACHINE ------------------------------------------
	type STATE_TYPE is
	(idle, setup_1, get_flash_incr1, latch_incr1, pause1, setup_2,  
	 get_flash_incr2, latch_incr2, update_sm_incr, do_command, get_sm_inc,  
	 chk_initial, bootup, incr_increment, store_increment, do_unlck_block, 
	 do_erase_block, backup, 
	 --setup_test_address, test_address, 
	 done);	
	signal CS, NS : STATE_TYPE;	
		
--========================================================
--========================================================
begin

-- FLASH_Addr_In(23 downto 1) is set in the state machine --

--converted_Addr(24 downto 1)	<= ('0' & FLASH_Addr_In(23 downto 1)) - X"040000";
converted_Addr(24 downto 1)	<= ('0' & FLASH_Addr_In(23 downto 1)); -- after July 26, 2017
FLASH_Addr(24 downto 1)	<= converted_Addr(24 downto 1);

--**-- --**--   --**--   --**--   --**--   --**--   --**--   
-- COMMAND SYNCHRONIZATION AND LATCHING--------------------------------  
SYNC_BOOTUP: led -- syncronize the signal to start boot from flash sequence.
port map(	sig_in  => Bootup_Trigger, 
			clk     => clk, 
			clrn    => '1', 
			pls_out => do_bootup); 

SYNC_BACKUP: led -- syncronize the signal to start backup to flash sequence.
port map(	sig_in  => Backup_Trigger, 
			clk     => clk, 
			clrn    => '1', 
			pls_out => do_backup); 
			
BOOTUP_COMMAND_LATCH: vdffe			
port map(	d    	=> '1',
			clk    	=> clk,
			clrn 	=> NOT clear_bootup,
			prn  	=> '1',
			ena  	=> do_bootup,
			q    	=> cmd_bootup);

BACKUP_COMMAND_LATCH: vdffe			
port map(	d    	=> '1',
			clk    	=> clk,
			clrn 	=> NOT clear_backup,
			prn  	=> '1',
			ena  	=> do_backup,
			q    	=> cmd_backup);
			
-- DATA REGISTER 1 ------------------------------------------------
DATA_REGISTER1: dffr
generic map(32)
port map( 	d    => UPD_Data_From_Mem, 
			clk  => clk, 
			clrn => clrn_data1, 
			prn  => '1', 
			ena  => get_incr_1, 
			q    => data_reg1_q 
		);
	
	incr_block1  <= data_reg1_q ;
	
-- DATA REGISTER 2 ------------------------------------------------
DATA_REGISTER2: dffr
generic map(32)
port map( 	d    => UPD_Data_From_Mem, 
			clk  => clk, 
			clrn => clrn_data2, 
			prn  => '1', 
			ena  => get_incr_2, 
			q    => data_reg2_q 
		);
	
	incr_block2  <= data_reg2_q ;

-- ############# ------------------------------------------------
INCREMENT_COUNTER:
process(clk, nreset, get_sm_incr, inc_sm_incr, set_sm_1, set_sm_2)
begin
	if   (nreset = '0') 	then sm_incr <= (others => '0');
	elsif rising_edge(clk) 	then
		if   (get_sm_incr = '1')   then sm_incr <= UPD_Data_From_Mem;
		elsif(set_sm_1    = '1')   then sm_incr <= incr_block1;
		elsif(set_sm_2    = '1')   then sm_incr <= incr_block2;
		elsif(inc_sm_incr = '1')   then sm_incr <= sm_incr + 1;
		end if;
	end if;
end process;

---------------------------------------------------------------------
LAST_STORE_BIT_LATCH:
process(clk, nreset, get_last_store, last_store)
begin
	if (nreset = '0') 				then last_store_bit <= '0';
	elsif rising_edge(clk) then
		if(get_last_store = '1')   	then last_store_bit <= last_store;
		end if;
	end if;
end process;

---------------------------------------------------------------------
LAST_RESTORE_BIT_LATCH:
process(clk, nreset, get_last_restore, last_restore)
begin
	if (nreset = '0') 				then last_restore_bit <= '0';
	elsif rising_edge(clk) then
		if(get_last_restore = '1')   	then last_restore_bit <= last_restore;
		end if;
	end if;
end process;

-----------------------------------------------------------------------------
PAUSE_TIMER: -- Timer used to generate setup time for FLASH writes
process (clk, inc_pause_tmr)
begin
	if rising_edge(clk) then
		if   (inc_pause_tmr = '0') then pause_tmr <= "000";
		else							pause_tmr <= pause_tmr + 1;
		end if;
	end if;
end process;


------------------------------------------------------------
--**--   --**--   --**--   --**--   --**--   --**--   --**--  
------------------------------------------------------------
-- STATE MACHINE  
----------------------------------------------------
SM_REG: process(clk, nreset)
begin
	if(nreset='0')				then CS <= idle;
	elsif(rising_edge(clk))		then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process(CS, do_bootup, do_backup, cmd_bootup, cmd_backup, incr_block1,
		incr_block2, pause_tmr, sm_incr, Flash_Transfer_Done, 
		Flash_Operation_Done, converted_Addr) 

begin

	-- Defaults ----------------------------
		Unlck_Block		<= '0';
		Erase_Block		<= '0';
		Write_Flash		<= '0';
		Read_Flash		<= '0';
				
		RAM_Addr_In		<= SHARED_MEMORY;	
		FLASH_Addr_In	<= FLASH_BLK1_FIRST;
		NWords			<= SHARED_MEMORY_NWORDS;	
		
		get_incr_1		<= '0';
		clrn_data1      <= '1';	
		
		get_incr_2		<= '0';	
		clrn_data2      <= '1';	
		
		inc_sm_incr 	<= '0';
		get_sm_incr		<= '0';
		set_sm_1		<= '0';
		set_sm_2		<= '0';
		
		inc_pause_tmr	<= '0';
		
		clear_bootup 	<= '0';
		clear_backup 	<= '0';	
				
		get_last_restore  	<= '0';
		get_last_store  	<= '0';
		
		last_restore		<= '0';
		last_store			<= '0';

		UPD_Addr_to_Mem		<= (others => '0');
		UPD_Data_to_Mem 	<= (others => '0');
		UPD_WR_Pulse 		<= '0';
		UPD_Bus_Ena			<= '0';

		Backup_Setup_Done	<= '0';
		Process_Done		<= '0';
		state					<= X"0";
	----------------------------------------	
	case CS is
		when idle =>																		state	<= X"1";
			clrn_data1      <= '0';	
			clrn_data2      <= '0';	

			if ((do_bootup = '1') OR (do_backup = '1')) then 	
													NS <= get_sm_inc;
			else                   			NS <= idle;
			end if;
			
		------------------------------------------------------------------
		-- We will examine the FLASH Storage Increment value stored in the
		-- Shared Memory.  If the value is either 0 or 1 then we are not 
		-- going to be concerned with the current Increment values stored
		-- in Blocks A and B of the FLASH.  This initial value of the RAM
		-- can be controlled with the FPGA *.mif memory initialization file.
-- ##### >>>> The production mif file will set the increment to something 
-- ##### >>>> other than 0 or 1 to cause a normal operational boot up.
		when get_sm_inc =>																state	<= X"2"; 				
			UPD_Addr_to_Mem	<= SM_FLASH_BLK_INCR; --std_logic_vector(23 downto 0) := X"013FFC";
			UPD_Bus_Ena		<= '1';
			get_sm_incr		<= '1';  -- latch the value into "sm_incr" counter
			
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "100") then	NS <= chk_initial;
			else									NS <= get_sm_inc;
			end if;
		
		when chk_initial =>																state	<= X"3";			
			UPD_Addr_to_Mem	<= SM_FLASH_BLK_INCR; --hold addr after latch
			UPD_Bus_Ena		<= '1';
			
			if   (sm_incr = X"00000000") then	NS <= do_command;
			elsif(sm_incr = X"00000001") then	NS <= do_command;
			else  										NS <= setup_1;               
			end if;               
			
		------------------------------------------------------------------	
		-- Get increment variables that are stored in the Flash memory
		-- and put them in the shared memory.
		when setup_1 => 																	state	<= X"4";          		
			RAM_Addr_In		<= SM_FLASH_INCR_REG;	--std_logic_vector(23 downto 0) := X"013FF8";
			FLASH_Addr_In	<= FLASH_BLK1_INCR;
			NWords			<= X"0001";			-- transfer the increment value in Flash to Shared Memory
			Read_Flash		<= '1';
			if (Flash_Transfer_Done = '1') then	NS <= get_flash_incr1;
			else											NS <= setup_1;
			end if;
		
		when get_flash_incr1 =>															state	<= X"5";				
			UPD_Addr_to_Mem	<= SM_FLASH_INCR_REG;
			UPD_Bus_Ena		<= '1';
			get_incr_1		<= '1'; -- increment from flash into datareg1
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "100") then		NS <= latch_incr1;
			else										NS <= get_flash_incr1;
			end if;
		
		when latch_incr1 =>						NS <= pause1;						state	<= X"6";
			UPD_Addr_to_Mem	<= SM_FLASH_INCR_REG; --hold addr after latch
			UPD_Bus_Ena		<= '1';
			
			-- Check if block increment is undefined (erased and not rewritten) --
			if   (incr_block1 = X"FFFFFFFF") then 	clrn_data1 <= '0'; --reset to zero
			else								  				clrn_data1 <= '1';
			end if;
			
		------------------------------------------------------------------	
		when pause1 =>								NS <= setup_2;						state	<= X"6";
		  
		------------------------------------------------------------------	
		when setup_2 =>           								state	<= X"6";
			RAM_Addr_In		<= SM_FLASH_INCR_REG;	
			FLASH_Addr_In	<= FLASH_BLK2_INCR;
			NWords			<= X"0001";		
			Read_Flash		<= '1';
			if (Flash_Transfer_Done = '1') then	NS <= get_flash_incr2;
			else											NS <= setup_2;
			end if;
		
		when get_flash_incr2 =>															state	<= X"7";
			UPD_Addr_to_Mem	<= SM_FLASH_INCR_REG;
			UPD_Bus_Ena		<= '1';
			get_incr_2		<= '1'; -- increment from flash into datareg1
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "100") then		NS <= latch_incr2;
			else										NS <= get_flash_incr2;
			end if;
		
		when latch_incr2 =>						NS <= update_sm_incr;			state	<= X"8";
			UPD_Addr_to_Mem	<= SM_FLASH_INCR_REG; --hold addr after latch
			UPD_Bus_Ena		<= '1';
			
			-- Check if block increment is undefined (erased and not rewritten) --
			if   (incr_block2 = X"FFFFFFFF") then 	clrn_data2 <= '0'; --reset to zero
			else								  				clrn_data2 <= '1';
			end if;
		
		when update_sm_incr =>					NS <= do_command;					state	<= X"8";
			if(incr_block1 > incr_block2) then 
				set_sm_1 <= '1';  -- sm_incr <= incr_block1;
			else
				set_sm_2 <= '1';  -- sm_incr <= incr_block2;
			end if;
		
		------------------------------------------------------------------	
		when do_command =>																state	<= X"9";
			-- now branch depending on the specific command --
			if   (cmd_bootup = '1')   then 	NS <= bootup;
         elsif(cmd_backup = '1')   then 	NS <= incr_increment;
			else                   			NS <= done;
			end if;
			
		------------------------------------------------------------------	
		when bootup => 																	state	<= X"A";
		--restore
			clear_bootup 	 <= '1'; -- clear the command to bootup
			get_last_restore <= '1';
			
			RAM_Addr_In		<= SHARED_MEMORY;	
			if   (sm_incr = X"00000000") then     ---- This may be an undefined restore
				FLASH_Addr_In	<= FLASH_BLK2_FIRST;
				last_restore 	<= '1'; 

			elsif(sm_incr = X"00000001") then
				FLASH_Addr_In	<= FLASH_BLK1_FIRST;
				last_restore 	<= '0'; 
              
			elsif(incr_block1 > incr_block2) then 
				FLASH_Addr_In	<= FLASH_BLK1_FIRST;
				last_restore 	<= '0'; 

			else
				FLASH_Addr_In	<= FLASH_BLK2_FIRST;
				last_restore 	<= '1'; 
					
			end if;
				
			Read_Flash		<= '1';
			-- NWords			<= SHARED_MEMORY_NWORDS; -- DEFAULT VALUE
			
			if (Flash_Transfer_Done = '1') then	NS <= done;
			else											NS <= bootup;
			end if;
			
		------------------------------------------------------------------
		------------------------------------------------------------------
		when incr_increment =>				NS <= store_increment;				state	<= X"B";
			inc_sm_incr <= '1';
			
		when store_increment =>															state	<= X"B";
			UPD_Addr_to_Mem	<= SM_FLASH_BLK_INCR;
			UPD_Data_to_Mem <= sm_incr;
			
			UPD_WR_Pulse 	<= '1';
			UPD_Bus_Ena		<= '1';
			
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "100") then		NS <= do_unlck_block;
			else										NS <= store_increment;
			end if;
		
		------------------------------------------------------------------	
		when do_unlck_block =>															state	<= X"C";
		
			Backup_Setup_Done <= '1';
		
			if   (sm_incr = X"00000001") 	then 		FLASH_Addr_In	<= FLASH_BLK1_FIRST;
			elsif(sm_incr = X"00000002") 	then 		FLASH_Addr_In	<= FLASH_BLK2_FIRST;
			elsif(incr_block1 > incr_block2)then 	FLASH_Addr_In	<= FLASH_BLK2_FIRST; 				
			else 								 				FLASH_Addr_In	<= FLASH_BLK1_FIRST;                														
			end if;               
			
			Unlck_Block 	<= '1';
			if   (Flash_Operation_Done = '1') then	NS <= do_erase_block;
			else												NS <= do_unlck_block;
			end if; 
			
		when do_erase_block =>															state	<= X"D";
			if   (sm_incr = X"00000001") 	then 		FLASH_Addr_In	<= FLASH_BLK1_FIRST;
			elsif(sm_incr = X"00000002") 	then 		FLASH_Addr_In	<= FLASH_BLK2_FIRST;
			elsif(incr_block1 > incr_block2)then 	FLASH_Addr_In	<= FLASH_BLK2_FIRST; 				
			else 								 				FLASH_Addr_In	<= FLASH_BLK1_FIRST;                														
			end if;               
			
			Erase_Block 	<= '1';
			if   (Flash_Operation_Done = '1') then	NS <= backup;
			else									NS <= do_erase_block;
			end if; 
		
		------------------------------------------------------------------
		when backup =>  																	state	<= X"E";
		-- store        	
			clear_backup 	<= '1';
			get_last_store  <= '1';
			
			RAM_Addr_In		<= SHARED_MEMORY;
			if   (sm_incr = X"00000001") 	then FLASH_Addr_In	<= FLASH_BLK1_FIRST;
												 last_store		<= '0';
			elsif(sm_incr = X"00000002") 	then FLASH_Addr_In	<= FLASH_BLK2_FIRST;
												 last_store		<= '1';
			elsif(incr_block1 > incr_block2)then FLASH_Addr_In	<= FLASH_BLK2_FIRST; 				
												 last_store		<= '1';
			else 								 FLASH_Addr_In	<= FLASH_BLK1_FIRST;                														
												 last_store		<= '0';
			end if;               
			
			Write_Flash		<= '1';
			if (Flash_Transfer_Done = '1') then	NS <= done;
			else											NS <= backup;
			end if;
			
		------------------------------------------------------------------
		when done =>						NS <= idle;									state	<= X"F";
			clear_backup 	<= '1';
			clear_bootup 	<= '1';
			Process_Done	<= '1';
			
		------------------------------------------------------------------
		when others =>              		NS <= idle;
	end case;
end process;

test_out <= get_incr_2;

end main;		
		
