Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 
USE work.DDS_VME_CONSTANTS.all;
-------------------------------------------------------------------------------
-- MODULE DESCRIPTION, June 24, 2010
--
-- The purpose of the Curve Request process is to poll all of the Request 
-- Curve and New Curve semaphores (flags) in Shared Memory.  When the host
-- processor wishes to read the values for a particular curve in the
-- DDS_VME module it will "set" the associated Request Semaphore for that
-- curve.  Once the DDS_VME module has detected this it will copy the values
-- in the curves memory to a common Curve Buffer in shared memory and "clear"
-- the Request Semaphore.  The host processor will be watching for the Request
-- Semaphore to be "cleared" and when it is the host processor will retrieve the
-- requested curve from the Curve Buffer.
--
-- When the host processor wishes to write a particular curve to the DDS-VME 
-- module it will "clear" the CTBFREE semaphore (Curve Table Buffer Free) and
-- will then proceed to write the new curve to the buffer.  It will then "set"
-- the New Curve Semaphore to signal the DDS-VME module which curve is to be
-- updated with the data in the Curve Buffer.  The host processor will also 
-- then release or "reset" the CTBFREE semaphore.  Once the DDS_VME has detected
-- the New Curve semaphore having been set it will "clear" the CTBFREE semaphore,
-- transfer the curve from the Curve Buffer to the particular curve in memory,
-- and then "reset" CTBFREE and "clear" the New Curve semaphore.

-- If CTBFREE is not found "set" when polling the process quits and will not
-- continue polling until triggered again.

----------------------------------------------------------------------------
-- VERSION 2: Made the timer pause_tmr clear to zero when not enabled.  This
--            timer is used to provide wait states for memory reads and writes.
--            some states in the state machine were then unnecessary and removed.

----------------------------------------------------------------------------
-- VERSION 3: This version no longer writes to the CURVE STATUS Word to indicate
-- 				that a new curve has been loaded and needs to be interpolated
--				and or transfered to the RAM curve memory.  NOW, a status
--				register in the curve interpolation process is pulsed to
--				indicated this.

entity Curve_Req_Process_v3 is 
	port( 
		clk					: in std_logic;
		nreset				: in std_logic;
		UPD_Data_From_Mem : in std_logic_vector(31 downto 0);
		Set_CTBFREE_Sem	: in std_logic;
		Curve_Req_Trigger : in std_logic;		

		UPD_Data_to_Mem	: out std_logic_vector(31 downto 0);
		UPD_Addr_to_Mem	: out std_logic_vector(23 downto 0);
		UPD_WR_Pulse		: out std_logic;
		UPD_RD_Pulse		: out std_logic;
		UPD_Bus_Ena			: out std_logic;
		Set_Interp_Status	: out std_logic_vector(8 downto 1);
		
		Curve_Changed		: out std_logic; -- signals need for backup
		Crv_Req_Proc_Done	: out std_logic  -- Strobes when done
			        
		);
	end Curve_Req_Process_v3;
	
architecture main of Curve_Req_Process_v3 is	

-- Set_Curve_Status signals -------------------------------------------
	signal curve_status_bits	: std_logic_vector(8 downto 1);

-- Tigger inputs ------------------------------------------------------
	signal do_crv_update			: std_logic;
	signal init_crv_buffer			: std_logic;

-- SEM_POINTER signals -------------------------------------------------
	signal sem_ptr_nrst		: std_logic; -- counter reset signal
	signal sem_ptr_inc		: std_logic; -- counter increment enable strobe
	signal sem_ptr_load		: std_logic;
	signal sem_ptr_step		: std_logic_vector(3 downto 0);
	signal sem_ptr_first	: std_logic_vector(19 downto 0);
	signal sem_ptr_last		: std_logic_vector(19 downto 0);
	signal sem_ptr_done		: std_logic; -- count end flag
	signal sem_ptr_out		: std_logic_vector(19 downto 0);
	
-- SOURCE_POINTER signals ------------------------------------------------
	signal src_ptr_nrst		: std_logic; -- counter reset signal
	signal src_ptr_inc		: std_logic; -- counter increment enable strobe
	signal src_ptr_load		: std_logic;
	signal src_ptr_step		: std_logic_vector(3 downto 0);
	signal src_ptr_first	: std_logic_vector(19 downto 0);
	signal src_ptr_last		: std_logic_vector(19 downto 0);
	--signal src_ptr_done	: std_logic; -- count end flag
	signal src_ptr_out		: std_logic_vector(19 downto 0);

-- DESTINATION_POINTER signals -------------------------------------------------
	signal dst_ptr_nrst		: std_logic; -- counter reset signal
	signal dst_ptr_inc		: std_logic; -- counter increment enable strobe
	signal dst_ptr_load		: std_logic;
	signal dst_ptr_step		: std_logic_vector(3 downto 0);
	signal dst_ptr_first		: std_logic_vector(19 downto 0);
	signal dst_ptr_last		: std_logic_vector(19 downto 0);
	--signal dst_ptr_done	: std_logic; -- count end flag
	signal dst_ptr_out		: std_logic_vector(19 downto 0);

-- CRV_POINTER signals -------------------------------------------------
	signal crv_ptr_nrst		: std_logic; -- counter reset signal
	signal crv_ptr_inc		: std_logic; -- counter increment enable strobe
	signal crv_ptr_load		: std_logic;
	signal crv_ptr_step		: std_logic_vector(15 downto 0);
	signal crv_ptr_first		: std_logic_vector(19 downto 0);
	signal crv_ptr_last		: std_logic_vector(19 downto 0);
	--signal crv_ptr_done	: std_logic; -- count end flag
	signal crv_ptr_out		: std_logic_vector(19 downto 0);

-- Data Register 1 --------------------------------------------------
	signal get_data1		: std_logic;
	signal data_reg1_d	: std_logic_vector(31 downto 0);
	signal dreg1			: std_logic_vector(31 downto 0);

-- Transfer Word Counter -------------------------------------------
	signal xfer_nrst	: std_logic;		
	signal xfer_inc		: std_logic;	
	signal xfer_done	: std_logic;
	
-- Pause Timer ------------------------------------
	signal pause_tmr 	: std_logic_vector(2 downto 0);
	signal inc_pause_tmr: std_logic;
  

-- POLLING_STATE_MACHINE High Level --------------------------------------
	type STATE_TYPE is
(
	init, pause, idle, init_crv_poll, read_reqsem, 
	check_reqsem, read_newsem, check_newsem, inc_crv_ptr, 
	xfer_crv2buf, inc_crv_ptr2, xfer_buf2crv, read_ctbfree,
	check_ctbfree, clear_ctbfree, do_transfer, read_source, 
	transfer_pause, write_dest, increment_ptrs, clear_sem, done
);		
 	signal CS, NS : STATE_TYPE;	

--########################################################
--========================================================
--						B E G I N
--========================================================
--########################################################

begin

--**-- --**--   --**--   --**--   --**--   --**--   --**-- 
-- STROBE SYNCHRONIZATION --------------------------------  
SYNC_UPDATE: led -- syncronize the signal to start update polling sequence.
port map(	sig_in  => Curve_Req_Trigger, 
				clk     => clk, 
				clrn    => '1', 
				pls_out => do_crv_update); 

SYNC_INIT: led -- syncronize the signal to initialize the update polling sequence.
port map(	sig_in  => Set_CTBFREE_Sem, 
				clk     => clk, 
				clrn    => '1', 
				pls_out => init_crv_buffer); 
	
-- ADDR_POINTER_COUNTER --------------------------------------------
SEM_ADDR_POINTER: countby_n -- Address Pointer into Curve Buffer Semaphores
generic map(20,4)
port map(	clk  	 => clk,
			clrn 	 => sem_ptr_nrst,
			ena  	 => sem_ptr_inc,
			load 	 => sem_ptr_load,
			ld_data  => sem_ptr_first,
			step	 => sem_ptr_step,
			flag_cnt => sem_ptr_last,
			flag_out => sem_ptr_done,
			cnt_out  => sem_ptr_out 
		);

-- ADDR_POINTER_COUNTER 2-------------------------------------------
SOURCE_ADDR_POINTER: countby_n -- Address Pointer to data transfer source
generic map(20,4)
port map(	clk  	 => clk,
			clrn 	 => src_ptr_nrst,
			ena  	 => src_ptr_inc,
			load 	 => src_ptr_load,
			step	 => src_ptr_step,
			ld_data  => src_ptr_first,
			flag_cnt => src_ptr_last,
			flag_out => open, --src_ptr_done,
			cnt_out  => src_ptr_out 
		);

-- ADDR_POINTER_COUNTER 3--------------------------------------------
DEST_ADDR_POINTER: countby_n -- Address Pointer to data transfer destination
generic map(20,4)
port map(	clk  	 => clk,
			clrn 	 => dst_ptr_nrst,
			ena  	 => dst_ptr_inc,
			load 	 => dst_ptr_load,
			ld_data  => dst_ptr_first,
			step	 => dst_ptr_step,
			flag_cnt => dst_ptr_last,
			flag_out => open, --dst_ptr_done,
			cnt_out  => dst_ptr_out 
		);
		
-- ADDR_POINTER_COUNTER 4--------------------------------------------
CURVE_ADDR_POINTER: countby_n -- Address Pointer to Dedicated Curve Buffers
generic map(20,16)
port map(	clk  	 => clk,
			clrn 	 => crv_ptr_nrst,
			ena  	 => crv_ptr_inc,
			load 	 => crv_ptr_load,
			ld_data  => crv_ptr_first,
			step	 => crv_ptr_step,
			flag_cnt => crv_ptr_last,
			flag_out => open, --crv_ptr_done,
			cnt_out  => crv_ptr_out 
		);

-- TRANSFER_COUNTER ------------------------------------------------
TRANSFER_COUNT: counter -- Counts words transfered
generic map(11)
port map(	clk  	 => clk,
			clrn 	 => xfer_nrst,
			ena  	 => xfer_inc,
			load 	 => '0',
			updn	 => '1',
			ld_data  => (others => '0'),
			flag_cnt => "11111111111",
			flag_out => xfer_done,
			cnt_out  => open 
		);

-- DATA REGISTER 1 ------------------------------------------------
DATA_REGISTER_1: dffr
generic map(32)
port map( 	d    => data_reg1_d, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_data1, 
			q    => dreg1 
		);
	
-----------------------------------------------------------------------------
PAUSE_TIMER: -- Timer used to generate setup time for FLASH writes
process (clk, inc_pause_tmr)
begin
	if rising_edge(clk) then
		if   (inc_pause_tmr = '0') then pause_tmr <= "000";
		elsif(inc_pause_tmr = '1') then pause_tmr <= pause_tmr + 1;
		end if;
	end if;
end process;

-----------------------------------------------------------------------------
SET_CURVE_STATUS_LOGIC:
process(sem_ptr_out)
begin
	case sem_ptr_out is
		when SM_NEW_TBL1_SEM(19 downto 0) => curve_status_bits <= "00000001";
		when SM_NEW_TBL2_SEM(19 downto 0) => curve_status_bits <= "00000010";
		when SM_NEW_TBL3_SEM(19 downto 0) => curve_status_bits <= "00000100";
		when SM_NEW_TBL4_SEM(19 downto 0) => curve_status_bits <= "00001000";
		when SM_NEW_TBL5_SEM(19 downto 0) => curve_status_bits <= "00010000";
		when SM_NEW_TBL6_SEM(19 downto 0) => curve_status_bits <= "00100000";
		when SM_NEW_TBL7_SEM(19 downto 0) => curve_status_bits <= "01000000";
		when SM_NEW_TBL8_SEM(19 downto 0) => curve_status_bits <= "10000000";
		when others =>						 curve_status_bits <= "00000000";
		-- NOTE: If the SEM pointer stopped on a REQ semaphore instead
		--       of a NEW semaphore, the curve_status_bits word will
		--       be "00000000" when written to the Curve Interpolation
		--       status register.
	end case;
end process;
	
--**--   --**--   --**--   --**--   --**--   --**--   --**--  
----------------------------------------------------
-- STATE MACHINE  
----------------------------------------------------
SM_REG: process(clk, nreset)
begin
	if(nreset='0')				then CS <= idle;
	elsif(rising_edge(clk))		then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process
(CS, do_crv_update, UPD_Data_From_Mem, dreg1, sem_ptr_out, pause_tmr, 
 crv_ptr_out, src_ptr_out, dst_ptr_out, xfer_done, sem_ptr_done, 
 init_crv_buffer, curve_status_bits )
begin
	-- Defaults -------------------------------------
		sem_ptr_nrst	<= '1';
		sem_ptr_inc		<= '0';
		sem_ptr_load	<= '0';
		sem_ptr_step	<= "0100";
		sem_ptr_first	<= SM_REQ_TBL1_SEM(19 downto 0);
		sem_ptr_last	<= SM_NEW_TBL8_SEM(19 downto 0);
		
		crv_ptr_nrst	<= '1';
		crv_ptr_inc		<= '0';
		crv_ptr_load	<= '0';
		crv_ptr_step	<= X"2000";
		crv_ptr_first	<= SM_CURVE_TBL1_STAT(19 downto 0);
		crv_ptr_last	<= SM_CURVE_TBL8_STAT(19 downto 0);

		src_ptr_nrst 	<= '1';
      src_ptr_inc 	<= '0';
      src_ptr_load 	<= '0';
		src_ptr_step	<= "0100";
      src_ptr_first	<= SM_CURVE_BUF_SIZE(19 downto 0);
		src_ptr_last	<= X"FFFFF";
		
		dst_ptr_nrst	<= '1';
		dst_ptr_inc		<= '0';
		dst_ptr_load	<= '0';
		dst_ptr_step	<= "0100";
		dst_ptr_first	<= SM_CURVE_BUF_SIZE(19 downto 0);
		dst_ptr_last	<= X"FFFFF";

		get_data1		<= '0';  -- Strobe to latch data into register 1
		data_reg1_d		<= UPD_Data_From_Mem;
		
		xfer_nrst		<= '1';	
		xfer_inc 		<= '0';	
                
		inc_pause_tmr	<= '0';
				
		UPD_Addr_to_Mem <= SM_CTBFREE_SEM;
		UPD_Data_to_Mem <= X"00000001";
		UPD_WR_Pulse 	<= '0';
		UPD_RD_Pulse 	<= '0';
		UPD_Bus_Ena		<= '1';  -- enabled except in the idle state
		
		Set_Interp_Status <= (others => '0');
		Curve_Changed		<= '0';
		Crv_Req_Proc_Done	<= '0';
				
	--------------------------------------------------

	case CS is
		--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	
		when idle =>
			UPD_Bus_Ena		<= '0'; -- enabled except in the idle state
			
			sem_ptr_nrst		<= '0';
			src_ptr_nrst		<= '0';
			dst_ptr_nrst		<= '0';
			crv_ptr_nrst		<= '0';
			xfer_nrst			<= '0';	

			
			if   (do_crv_update = '1')   then 	NS <= init_crv_poll;
			elsif(init_crv_buffer = '1') then	NS <= init;
			else                   					NS <= idle;
			end if;
		
		when init => 									
		-- set_ctbfree => 				
			UPD_Addr_to_Mem <= SM_CTBFREE_SEM;
			UPD_Data_to_Mem <= X"00000001";
			UPD_WR_Pulse <= '1'; -- << The write operation
			
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "011") then		NS <= pause;
			else										NS <= init;
			end if;
			
		when pause =>						NS <= idle;	
		-- we also come here if the test of ctbfree below fails !!!
			UPD_Addr_to_Mem <= SM_CTBFREE_SEM;
			UPD_Data_to_Mem <= X"00000001"; -- Addr and Data Held-over
			
			Crv_Req_Proc_Done <= '1';
			
		-------------------------------------------------------
		when init_crv_poll => 				NS <= read_reqsem;
			sem_ptr_load	<= '1';
			sem_ptr_first	<= SM_REQ_TBL1_SEM(19 downto 0);
			
			crv_ptr_load	<= '1';
			crv_ptr_first	<= SM_CURVE_TBL1_STAT(19 downto 0);

		----------------------------------------------------------------
		-- start loop checking curve request and new curve semaphores --
		----------------------------------------------------------------
		when read_reqsem =>				
			UPD_Addr_to_Mem <= "0000" & sem_ptr_out(19 downto 0);
			get_data1		<= '1';
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (sem_ptr_done = '1')  then	NS <= done;
			elsif(pause_tmr = "011") then	NS <= check_reqsem;
			else							NS <= read_reqsem;
			end if;
			
		when check_reqsem =>					
			UPD_Addr_to_Mem <= "0000" & sem_ptr_out(19 downto 0); -- held-over
			if(dreg1 > X"00000000")  then	NS <= inc_crv_ptr;
				-- a request to read a curve has been made				
			else								
											NS <= read_newsem;
				sem_ptr_inc 	<= '1'; -- step to new table semaphore
				sem_ptr_step	<= "0100";	-- step 4 bytes      			                     
			end if;
			
		---------------------------------------------------------
		when read_newsem =>				
			UPD_Addr_to_Mem <= "0000" & sem_ptr_out(19 downto 0);
			get_data1		<= '1';
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if  (pause_tmr = "011") then	NS <= check_newsem;
			else									NS <= read_newsem;
			end if;
			
		when check_newsem =>					
			UPD_Addr_to_Mem <= "0000" & sem_ptr_out(19 downto 0); -- held-over
			if(dreg1 > X"00000000")  then	NS <= inc_crv_ptr2;
				-- a new curve table is available				
			else								
											NS <= read_reqsem;
				sem_ptr_inc 	<= '1';  	-- step to request table sem
				sem_ptr_step	<= "0100";
					      
				crv_ptr_inc 	<= '1';  	-- step to next curve table
				crv_ptr_step	<= X"2000";	      

			end if;
			
	--------------------------------------------------------------
	-- end loop checking curve request and new curve semaphores --
	--------------------------------------------------------------

	---------------------------------------------------------
	-- We will transfer the data in the requested curves curve table
	-- to the curve buffer.		
		when inc_crv_ptr =>					NS <= xfer_crv2buf;
			crv_ptr_inc 	<= '1'; -- step past status word 
			crv_ptr_step	<= X"0004";	      			                     
		
		when xfer_crv2buf =>				NS <= read_ctbfree;	
			-- source = curve table
			src_ptr_load 	<= '1';
			src_ptr_first	<= crv_ptr_out(19 downto 0); 
			
			-- destination = curve buffer
			dst_ptr_load	<= '1';
			dst_ptr_first	<= SM_CURVE_BUF_SIZE(19 downto 0);
		
		----------------------------------------------------------------
		-- We will transfer the data in the buffer to the requested 
		-- curves curve table.  We first set the curves status word
		-- to indicate that the curve is new and needs to be transfered
		-- and possibly interpolated into the associated curve table
		-- RAM.	
		when inc_crv_ptr2 =>					NS <= xfer_buf2crv;
			crv_ptr_inc 	<= '1'; -- step past status word 
			crv_ptr_step	<= X"0004";	      			                     

		when xfer_buf2crv =>				NS <= read_ctbfree;	
			-- source = curve buffer
			src_ptr_load 	<= '1';
			src_ptr_first	<= SM_CURVE_BUF_SIZE(19 downto 0);
			
			-- destination = curve table
			dst_ptr_load	<= '1';
			dst_ptr_first	<= crv_ptr_out(19 downto 0); 
			
			-- Indicate that a curve has been updated (or will be soon).
			Curve_Changed		<= '1';
		
		----------------------------------------------------------------
		----------------------------------------------------------------
		when read_ctbfree =>				
			UPD_Addr_to_Mem <= SM_CTBFREE_SEM;
			get_data1		<= '1';
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (pause_tmr = "011") then	NS <= check_ctbfree;
			else									NS <= read_ctbfree;
			end if;
			
		when check_ctbfree =>					
			UPD_Addr_to_Mem <= SM_CTBFREE_SEM; -- held-over
			if(dreg1 > X"00000000")  then	NS <= clear_ctbfree;-- buffer is available
			else									NS <= pause; -- aborts the polling
			end if;										 		 -- pause is just before idle
			
		----------------------------------------------------------------
		when clear_ctbfree =>				
		-- clear the CTBFREE Semaphore ---------
			UPD_Addr_to_Mem <= SM_CTBFREE_SEM;
			UPD_Data_to_Mem <= X"00000000";
			UPD_WR_Pulse <= '1'; -- << The write operation
			
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "011") then	NS <= do_transfer;
			else									NS <= clear_ctbfree;
			end if;
		
		when do_transfer =>					NS <= read_source;
			UPD_Addr_to_Mem <= SM_CTBFREE_SEM;
			UPD_Data_to_Mem <= X"00000000"; -- Addr and Data are held-over
			
			xfer_nrst		<= '0';	-- reset transfer count
			
-- start data transfer loop --
		when read_source =>					
			UPD_Addr_to_Mem <= "0000" & src_ptr_out(19 downto 0); 
			get_data1		<= '1'; -- latches data into dreg1
																	
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if  (pause_tmr = "011") then	NS <= transfer_pause;
			else							NS <= read_source;
			end if;
		
		when transfer_pause =>				NS <= write_dest;			
			UPD_Addr_to_Mem <= "0000" & src_ptr_out(19 downto 0);
			-- this state allows the pause_tmr to reset --
			
		when write_dest =>					
			UPD_Addr_to_Mem	<= "0000" & dst_ptr_out(19 downto 0);
			UPD_Data_to_Mem <= dreg1;			
			UPD_WR_Pulse 	<= '1';
			
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "011") then	NS <= increment_ptrs;
			else									NS <= write_dest;
			end if;
			
		when increment_ptrs =>				
			UPD_Addr_to_Mem	<= "0000" & dst_ptr_out(19 downto 0);
			UPD_Data_to_Mem <= dreg1;	-- Addr and Data busses held over		
			
			if   (xfer_done = '1') then	NS <= clear_sem;
			else							
													NS <= read_source;
				xfer_inc 		<= '1';	
				src_ptr_inc 	<= '1';
				dst_ptr_inc		<= '1';
			end if;
		                 	
-- end data transfer loop --
		------------------------------------------------------------
				
		when clear_sem =>						

		-- clear the Semaphore just serviced ---------
			UPD_Addr_to_Mem <= "0000" & sem_ptr_out(19 downto 0);
			UPD_Data_to_Mem <= X"00000000";
			UPD_WR_Pulse <= '1'; -- << The write operation
			
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "011") then		NS <= done;
			else										NS <= clear_sem;
			end if;					
							
		-- set status register in the interpolation process.
		-- if this was not a new curve then curve_status_bits
		-- will be all zeros and status will not change.
		Set_Interp_Status <= curve_status_bits;

			
		when done =>						NS <= init;					
			--this state allows pause_tmr to reset--
			-- GO TO INIT TO SET CTBFREE AGAIN
			
		when others =>              		NS <= idle;
	end case;
end process;
	
end;