Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 
USE work.DDS_VME_CONSTANTS.all;

----------------------------------------------------------------
-- Placement of these 32 bit registers is controlled by the 
-- constant VME_CTRL_UPPER_FIRST.  This constant is defined in
-- the included file DDS_VME_CONSTANTS.vhd.

entity Debug_Data_Select is 
	port( 	
			clk				: in std_logic;
			nreset			: in std_logic;
			Addr_to_Mem		: in std_logic_vector(23 downto 0);
			Read_Pulse_In	: in std_logic;
			
			Start_Addr		: in std_logic_vector(23 downto 0);
			Enable_Capture	: in std_logic;
						
			Frequency		: in std_logic_vector(31 downto 0);
			Freq_Strobe		: in std_logic;
			
			Phase_A			: in std_logic_vector(15 downto 0);
			Phase_A_Strobe	: in std_logic;
			
			Phase_B			: in std_logic_vector(15 downto 0);
			Phase_B_Strobe	: in std_logic;
			
			End_Cycle		: in std_logic;
			FIFO_FULL		: in std_logic;
						
			Data_to_FIFO 	: out std_logic_vector(31 downto 0);
			WR_FIFO			: out std_logic
			
		);
	end Debug_Data_Select;
	
architecture main of Debug_Data_Select is	

signal addr_in		: std_logic_vector(23 downto 0);
signal start		: std_logic;
signal stop			: std_logic;
signal data_out		: std_logic_vector(31 downto 0);
signal strobe_out	: std_logic;

-- SETUP_STATE_MACHINE ---------------------------------------
type STATE_TYPE is
(idle, wait1, run, hold);
signal CS, NS : STATE_TYPE;
            
begin

SYNC_START: led -- syncronize the signal to start update polling sequence.
port map(	sig_in  => Enable_Capture, 
			clk     => clk, 
			clrn    => '1', 
			pls_out => start); 

SYNC_STOP: led -- syncronize the signal to stop update polling sequence.
port map(	sig_in  => End_Cycle, 
			clk     => clk, 
			clrn    => '1', 
			pls_out => stop); 

ADDRESS_IN_REG:
process(clk, nreset, Read_Pulse_In)
begin
	if   (nreset = '0') then addr_in <= (others => '0');
	elsif rising_edge(clk) then
		if   (Read_Pulse_In = '1') then addr_in <= Addr_to_Mem;
		end if;
	end if;
end process;

DATA_SELECT:
process( Addr_to_Mem, Start_Addr, Frequency, Freq_Strobe,
		 Phase_A, Phase_A_Strobe, Phase_B, Phase_B_Strobe)
begin
	if   (Start_Addr >= FREQ_CURVE_ADDR) AND 
		 (Start_Addr < FREQ_CURVE_ADDR + X"009C3C" ) then 
			data_out   <= Frequency;
			strobe_out <= Freq_Strobe;
			
	elsif(Start_Addr >= POFF1_CURVE_ADDR) AND 
		 (Start_Addr <  POFF1_CURVE_ADDR + X"004E1E" ) then 
			data_out   <= X"0000" & Phase_A;			
			strobe_out <= Phase_A_Strobe;
			             
	elsif(Start_Addr >= POFF2_CURVE_ADDR) AND 
		 (Start_Addr <  POFF2_CURVE_ADDR + X"004E1E" ) then 
			data_out   <= X"0000" & Phase_B;			
			strobe_out <= Phase_B_Strobe;
	else
			data_out   <= X"00000000";			
			strobe_out <= '0';	
	end if;

end process;

			
--{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{
--                S T A T E   M A C H I N E
--}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}
SM_REG: process(clk, nreset)
begin
	if(nreset='0')			then CS <= idle;
	elsif(rising_edge(clk))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process
(CS, start, stop, FIFO_FULL, addr_in, Start_Addr, data_out, strobe_out )
begin
	-- defaults ----------------------
		Data_to_FIFO <= X"bad8add9";
		WR_FIFO		 <= '0';
	
	case CS is
		when idle =>	
			if   (start = '1') then				NS <= wait1;
			else								NS <= idle;
			end if;

		when wait1 =>
			if   (stop = '1') then				 NS <= idle;
			elsif(addr_in = Start_Addr) then 	 NS <= run;
			else								 NS <= wait1;
			end if;		
	        
		when run =>	    
			if   (stop = '1') then				NS <= idle;
			elsif(FIFO_FULL = '1') then			NS <= hold;
			else								NS <= run;
				Data_to_FIFO <= data_out;
				WR_FIFO		 <= strobe_out;
			end if;
		
		when hold =>
			if   (stop = '1') then				NS <= idle;
			elsif(FIFO_FULL = '1') then			NS <= hold;
			else								NS <= idle;
			end if;		
		
		when others =>							NS <= idle;
	end case;
end process;
end;