Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

USE work.pkg_081006.all;
USE work.DDS_VME_CONSTANTS.all;

---------------------------------------------------------
-- version 3 adds the phase error trim

-- Version 8: In version 8 the trimmed and gated phase error is 
-- synchronized to the Update_Pulse before being set into the 
-- multiplication with the phase error gain and addition to then
-- frequency curve value.  Previously, the beam gate would go away
-- and the gated phase error value sometimes went to zero right in
-- the middle of the multiplication and the frequency value would
-- get the corrupted result.

-- Version 9: In version 9 the Update_Pulse is always applied to
-- keep firing the DDS update strobe so that phase adjustments due
-- to RPOS phase drive would continue smoothly from cycle to cycle.
-- Hence, the previous inputs Initialize and Update_Enable were 
-- eliminated.

entity Frequency_Feedback_v9 is
port(
		clk						: in std_logic;
		nreset					: in std_logic;
		Update_Pulse			: in std_logic;
		Freq_Curve				: in std_logic_vector(31 downto 0);
		
		Error_Gain				: in std_logic_vector(15 downto 0);
		Phase_Err		 		: in std_logic_vector(15 downto 0);
		Phase_Err_Strb			: in std_logic;
		Phase_Err_Enable		: in std_logic;
		Phase_Err_Trim_OB		: in std_logic_vector(15 downto 0); -- Offset binary 0x8000 = zero
		
		TransFreqThresh		: in std_logic_vector(31 downto 0);

		Freq_Curve_Out			: out std_logic_vector(31 downto 0);
		Base_Freq_Out			: out std_logic_vector(31 downto 0);
		BPM_LO_Out				: out std_logic_vector(31 downto 0);
		DDS_Update_Strobe		: out std_logic;
		freq_error_out			: out std_logic_vector(31 downto 0);
		trim_cout				: out std_logic;
		Trans_Gate_Out			: out std_logic
);
end Frequency_Feedback_v9;

--########################################################
--				A R C H I T E C T U R E
--########################################################
architecture main of Frequency_Feedback_v9 is
-- Synchronized Inputs --------------------------------
	signal gated_phase_err	: std_logic_vector(15 downto 0);
	
-- The Error input is double buffered
	signal PhErr_in		: std_logic_vector(15 downto 0);
	signal PHErr_Buff		: std_logic_vector(15 downto 0);
	
	signal PHErr_trim_2s : std_logic_vector(15 downto 0);
	signal trimmed_pherr	: std_logic_vector(15 downto 0);
	
	signal freq_error 	: std_logic_vector(31 downto 0);
	
	signal delay		: std_logic;

-- Altera Math Functions ------------------------------
component signed_sub32
	PORT
	(
		dataa		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		overflow	: OUT STD_LOGIC ;
		result		: OUT STD_LOGIC_VECTOR (31 DOWNTO 0)
	);
end component;

component addsub32
	PORT
	(
		add_sub		: IN STD_LOGIC ;
		clken			: IN STD_LOGIC ;
		clock			: IN STD_LOGIC ;
		dataa			: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		datab			: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		cout			: OUT STD_LOGIC ;
		overflow		: OUT STD_LOGIC ;
		result		: OUT STD_LOGIC_VECTOR (31 DOWNTO 0)
	);
end component;

	signal ovf1			: std_logic;
	signal sub_result	: std_logic_vector(31 downto 0);
	signal ovf2			: std_logic;
	signal bpm_lo		: std_logic_vector(31 downto 0);
	
component multiply_by_gain
	PORT
	(
		clock		: IN STD_LOGIC ;
		dataa		: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		result		: OUT STD_LOGIC_VECTOR (31 DOWNTO 0)
	);
end component;	           

component signed_adsub_16
	PORT
	(
		add_sub	: IN STD_LOGIC ;
		dataa		: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		cout		: OUT STD_LOGIC ;
		result	: OUT STD_LOGIC_VECTOR (15 DOWNTO 0)
	);
end component;

--========================================================
--						B E G I N
--========================================================
begin

--update <= (Update_Pulse AND Update_Enable) OR Initialize;

PHErr_trim_2s <= not Phase_Err_Trim_OB(15) & Phase_Err_Trim_OB(14 downto 0);

DELAYED_STROBE_OUT:
process(clk, nreset, Update_Pulse, delay)
begin
	if(nreset='0')			then 
		delay					<= '0';
		DDS_Update_Strobe	<= '0';
	elsif(rising_edge(clk))	then 
		delay   				<= Update_Pulse;
		DDS_Update_Strobe	<= delay;
	end if;
end process;

----------------------------------------------------------
PHASE_ERROR_REGISTER:
process(clk, nreset, Phase_Err, Phase_Err_Strb)
begin
	if   (nreset = '0') then			 PhErr_in <= (others => '0');			 	
	elsif rising_edge(clk) then		
		if	(Phase_Err_Strb = '1') then PhErr_in <= Phase_Err;
		end if;
	end if;
end process;

TRIM_PHASE_ERROR: 
signed_adsub_16 port map
	(
		add_sub		=> '1',
		dataa			=> PhErr_in,
		datab			=> PHErr_trim_2s,
		cout			=>	trim_cout,
		result		=> trimmed_pherr
	);		

BEAM_PRESENT_ERROR_CONTROL_Version_2:
process(trimmed_pherr, Phase_Err_Enable)
begin
	if (Phase_Err_Enable = '1') then gated_phase_err <= trimmed_pherr;
	else                             gated_phase_err <= (others => '0');
	end if;
end process;

BUFFERS:
process(clk, nreset, Update_Pulse, gated_phase_err, Freq_Curve) 
begin
	if   (nreset = '0') then
			Base_Freq_Out	<= (others => '0');
			PHErr_Buff		<= (others => '0');	
		
	elsif rising_edge(clk) then		
		if  (Update_Pulse = '1') then
				-- syncronizes the base freq with the feedback version
			Base_Freq_Out	<= Freq_Curve; 
			PHErr_Buff		<= gated_phase_err;	
		end if;
		
	end if;
end process;

------------------------------------------------------------
ERROR_GAIN_MULTIPLICATION:
multiply_by_gain port map( clock 	=> clk,
									dataa		=> PHErr_Buff,
									datab		=> Error_Gain,
									result	=> freq_error  );
							
freq_error_out <= freq_error;

FREQ_FEEDBACK: 
signed_sub32 port map(	dataa		=>	Freq_Curve,
								datab		=>	freq_error,
								overflow => ovf1,
								result	=> sub_result );
							
-- OUTPUT --------------------------------
--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Freq_Curve_Out <= sub_result;

--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

process(clk, nreset, TransFreqThresh, sub_result)
begin
	if   (nreset = '0') then Trans_Gate_Out <= '0';
	elsif (rising_edge(clk)) then
		if (sub_result > TransFreqThresh) then Trans_Gate_Out <= '1';
		else												Trans_Gate_Out <= '0';
		end if;
	end if;
end process;
--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

BPM_LOCAL_OSC: 
addsub32
	port map
	(
		add_sub	=> '1',
		clken		=> '1',
		clock		=> clk,
		dataa		=> sub_result,
		datab		=> BPM_LO_FREQ, -- set in dds_vme_constants.vhd
		--cout		=> NULL,
		overflow	=> ovf2,
		result	=> bpm_lo
	);

BPM_LO_Out <= bpm_lo;
--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

end;             