Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 

------------------------------------------------------
-- MODULE DESCRIPTION
-- 

--  TODO:
--  ADD A VME PROCESS TIMEOUT THAT WILL RESET THIS INTERFACE.

-- May 28, 2010 Version 5
--		In version 5 we changed the Interface Ready input
--		logic to a Transfer Done type of logic.


------------------------------------------------------

entity VME_Interface_v5 is 
	port( 
			clk				: in std_logic;
			nreset			: in std_logic;
			Mem_Data_To_VME	: in std_logic_vector(31 downto 0);
			VME_Data_In_lsw : in std_logic_vector(15 downto 0);
			vmeAddr			: in std_logic_vector(23 downto 1);
			Board_Addr		: in std_logic_vector( 3 downto 0);
			AM				: in std_logic_vector( 5 downto 0);
			IACK_n			: in std_logic;
			LWORD_n			: in std_logic;
			AS_n			: in std_logic;
			DS0_n			: in std_logic;
			DS1_n			: in std_logic;
			WRITE_n			: in std_logic;
						
			HoldOff_VME		: in std_logic;
			
			vmeExtAddr		: in std_logic_vector(25 downto 24);
			Ext_Board_Select: in std_logic;
			Transfer_Done 	: in std_logic;
			
			VME_Data_To_Mem	: out std_logic_vector(31 downto 0);
			VME_Addr_To_Mem : out std_logic_vector(25 downto 0);
			VME_WR_Pulse    : out std_logic;
			VME_Write_Active: out std_logic; -- used by InterFGPA logic
			VME_WR_Lower_Mem: out std_logic; -- used by InterFGPA logic					
			VME_RD_Lower_Mem: out std_logic; -- used by InterFGPA logic
			
			Upper_VME_Active_n	: out std_logic; -- used by VME bus transceivers
			Upper_Write_nRead 	: out std_logic; -- used by VME bus transceivers
			Lower_VME_Active_n 	: out std_logic;
			Lower_Write_nRead  	: out std_logic;
			DTACK				: out std_logic;
			BERR				: out std_logic;
			
			BS_Out			: out std_logic
			
		);
	end VME_Interface_v5 ;
	
architecture main of VME_Interface_v5  is

-- Syncronized input variables -------------------------------------------
signal  AS				: std_logic;
signal	DS				: std_logic;
signal	WR				: std_logic;
signal  IACK			: std_logic;

signal  sync_BS_d		: std_logic_vector(1 downto 0);
signal  sync_BS_q		: std_logic_vector(1 downto 0);
signal  BS			: std_logic;

-- Variables for combinatorial logic for board addressing ----------------
signal  Addr_Match		: std_logic;
signal  full_vme_addr	: std_logic_vector(25 downto 0);

signal	Block_Transfer	: std_logic;
signal	Program_Access  : std_logic;
signal  Data_Access		: std_logic;

signal  Standard_Address    : std_logic;

signal	Ext_Block_Transfer	: std_logic;
signal	Ext_Program_Access  : std_logic;
signal  Ext_Data_Access		: std_logic;

signal  Extended_Address    : std_logic;

signal  Block_xfer	 		: std_logic; -- syncronized version

signal  A32_nA24			: std_logic;
signal  D32_nD16			: std_logic;

-- Data Transfer Control Signals -----------------------------------------
signal Drive_VME		: std_logic;
signal WR_mem			: std_logic;
signal RD_mem			: std_logic;

signal VME_Active_n		: std_logic;
signal Write_nRead		: std_logic;
       
-- Address Latch ---------------------------------------------------------
signal LTCH_ADDR			: std_logic;
signal Lower_Memory_Addressed : std_logic;
signal VME_Latched_Addr		: std_logic_vector(25 downto 0);

-- Data Latches ----------------------------------------------------------
signal Latched_VME_Dout		: std_logic_vector(31 downto 16);
signal Latched_VME_Din		: std_logic_vector(31 downto 0);
signal Latch_VME_Dout		: std_logic;
signal Latch_VME_Din		: std_logic;

-- SETUP_STATE_MACHINE ---------------------------------------------------
type STATE_TYPE is
	(   idle, check_address, latch_address, data_setup_dly, decode_adr,
	         
	 set_lower_data, get_lower_data, latch_lower_data, get_upper_data,   
	 
	 latch_vme_data, set_upper_data, setup_data_pause, end_cycle, 
	 
	 strobe_wr, dtacken, xfer_done);	

	  
signal CS, NS : STATE_TYPE;
	
--========================================================================
--========================================================================
--                    B E G I N 
--========================================================================
--========================================================================	
begin
--========================================================================			
-- Board Select Logic 

-- STANDARD ADDRESS MODIFIER DECODE ---------------------------------------
-- HEX Value of AM(5..0)	|  Function
--      3F					|  Standard supervisory block transfer
--		3E					|  Standard supervisory program access
--		3D					|  Standard supervisory data access
--		3C					|  Reserved
--      3B					|  Standard nonprivileged block transfer
--		3A					|  Standard nonprivileged program access
--		39					|  Standard nonprivileged data access
--		38					|  Reserved
--
-- "Standard" implies that we are using address bits vmeAddr(23..1).
--
--      0F					|  Extended supervisory block transfer
--		0E					|  Extended supervisory program access
--		0D					|  Extended supervisory data access
--		0C					|  Reserved
--      0B					|  Extended nonprivileged block transfer
--		0A					|  Extended nonprivileged program access
--		09					|  Extended nonprivileged data access
--		08					|  Reserved
--
-- "Extended" implies that we are using address bits vmeAddr(31..1).



Block_Transfer	<='1' when AM = X"3F" OR AM = X"3B" else '0';
Program_Access  <='1' when AM = X"3E" OR AM = X"3A" else '0';
Data_Access		<='1' when AM = X"3D" OR AM = X"39" else '0';

Standard_Address <= Block_Transfer OR Program_Access OR Data_Access;

Ext_Block_Transfer	<='1' when AM = X"0F" OR AM = X"0B" else '0';
Ext_Program_Access  <='1' when AM = X"0E" OR AM = X"0A" else '0';
Ext_Data_Access		<='1' when AM = X"0D" OR AM = X"09" else '0';

Extended_Address <= Ext_Block_Transfer OR Ext_Program_Access OR Ext_Data_Access;

-------------------------------------------------------------------------------
BOARD_ADDRESS_MATCH: 
	process(vmeAddr, Board_Addr, HoldOff_VME, Standard_Address, Extended_Address, 
			Ext_Board_Select, vmeExtAddr)
	begin
	if (Standard_Address = '1') then
		full_vme_addr	<= "00" & vmeAddr(23 downto 1) & '0'; -- ignore A[25..24]
		if (vmeAddr(23 downto 20) = Board_Addr(3 downto 0)) then Addr_Match <='1'; 
		else 													 Addr_Match <='0';
		end if;
	elsif (Extended_Address = '1') then
		full_vme_addr	<= vmeExtAddr(25 downto 24) & vmeAddr(23 downto 1) & '0';
		if (Ext_Board_Select = '1') 						then Addr_Match <='1'; 
		else 													 Addr_Match <='0';
		end if;
	else
		full_vme_addr	<= "00" & vmeAddr(23 downto 1) & '0'; -- ignore A[25..24]
		Addr_Match <= '0';
	end if;
	end process;

	 
SYNCRONIZE_VME_INPUTS:	process(clk, nreset, DS0_n, DS1_n, WRITE_n)
	begin
		if(nreset = '0')  then 
			AS		<= '0';
			DS 		<= '0';
			WR 		<= '1';
			IACK	<= '0';
			BS		<= '0';
		elsif rising_edge(clk) then
			AS		<=  NOT AS_n;
			DS 		<= (NOT DS0_n) AND (NOT DS1_n); -- allow only even byte access.
			WR 		<=  NOT WRITE_n;
			IACK	<=  NOT IACK_n;	
			BS		<= ((NOT IACK) AND (NOT AS_n) AND Addr_Match); 
						
			if   (full_vme_addr(25 downto 0) < ("00" & X"020000") ) then
													Lower_Memory_Addressed <= '0';
			elsif(full_vme_addr(25 downto 0) < ("00" & X"080000") ) then
													Lower_Memory_Addressed <= '1';
			else									Lower_Memory_Addressed <= '0';
			end if;
		end if;
	end process;

SYNCRONIZE_BLOCK_XFER: process(clk, nreset, BS, Block_Transfer, Ext_Block_Transfer)	
	begin
		if(nreset = '0')  then 
			Block_xfer <= '0';
		elsif rising_edge(clk) then
			Block_xfer <= BS AND (Block_Transfer OR Ext_Block_Transfer);
		end if;
	end process;		
		

--=======================================================================
--=======================================================================
-- STATE MACHINE  
-------------------------------------------------------------------------
SM_REG: 
process(clk, nreset)
begin
	if(nreset='0')			then CS <= idle;
	elsif(rising_edge(clk))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process(CS, AS, BS, DS, WR, HoldOff_VME, Block_xfer, 
		Transfer_Done, VME_Latched_Addr, Lower_Memory_Addressed,
		VME_Data_In_lsw, Latched_VME_Dout, Latched_VME_Din)
		
begin
	-- DEFAULT OUTPUT VALUES --------------------------
	LTCH_ADDR 			<= '0';
	DTACK 				<= '0';
	BERR  				<= '0';
	
	VME_Addr_To_Mem(25 downto 0) <= VME_Latched_Addr(25 downto 0);
	VME_Data_To_Mem(31 downto 0) <= X"00000000";
	
	Latch_VME_Din  <= '0';
	Latch_VME_Dout <= '0';
	
	VME_WR_Pulse <= '0';
	
	WR_mem <= '0';
	RD_mem <= '0';
	
	Drive_VME <= '1'; 
	

	-------------------------------------------------------------------------
    
	case CS is
	-------------------------------------------------------------------------
	--                    INITIAL STATE
		when idle =>
			Drive_VME <= '0';
		
			if (HoldOff_VME  = '1') then 	NS <= idle;
			elsif (AS = '1') 		then 		NS <= check_address;
			else                   			NS <= idle;
			end if;
	-------------------------------------------------------------------------
	--              VME ADDRESS LATCHED IN UPPER FPGA
			
		when check_address => 
			Drive_VME <= '0';
			-- This state allows examination of the Board Select (AMATCH) bits.
			LTCH_ADDR     <= '1'; -- this is a DFF enable.
			
			if    (AS = '0')      	then 	NS <= idle; --just a glitch
			elsif (BS = '1')      	then	NS <= latch_address;
			else  								NS <= idle; 
			end if;
				
		when latch_address => 				
			-- All of the address bits vmeA[25..1] from both upper and lower are 
			-- latched in the Upper FPGA.
			LTCH_ADDR     <= '1'; -- this is a DFF enable.

			if    (BS = '0')      		then 	NS <= idle; --just a glitch
			elsif (HoldOff_VME  = '1')  then NS <= latch_address;
			elsif (DS = '1')      		then	NS <= decode_adr; 
			else  									NS <= latch_address;
			end if;
			
		when decode_adr =>
			
			if    (Lower_Memory_Addressed = '1') then 
				if (WR = '1') 		then    	NS <= set_lower_data; -- set LSW into lower fpga
				else								NS <= get_lower_data; -- get LSW from lower fpga
				end if;
			else  								NS <= data_setup_dly;-- Upper memory is being addressed.	
			end if;
			
		when data_setup_dly => 					
			-- Need this delay for memory data to setup.
			
			Latch_VME_Dout <= '1'; -- latch local data for possible read operation
			
			if (WR = '1') 		then    		NS <= get_upper_data; -- get MSW from lower fpga
			else									NS <= set_upper_data; -- set MSW into lower fpga
			end if;
				

	-------------------------------------------------------------------------------	
	--        WRITE VME ADDRESS BITS AND DATA BITS [15..0] TO LOWER FPGA								  
	-- 	Lower VME Write --				
	-- Now we will drive address bit from the upper fpga and latch them
	-- in the lower fpga.  We also write the least significant word of 
	-- VME data (15 .. 0) to the Lower fpga.
		
	-- RECALL: Settting Addr[19] = '1' signals the Lower FPGA that a VME Write
	--         operation is active.  This routes VME_Data_Out(31..16) 
	--         to the highest memory data bits(31..16) attached to the
	--		   Lower FPGA.	
	
		when set_lower_data =>				
			VME_Addr_To_Mem(25 downto 0) <= "0000001" & VME_Latched_Addr(18 downto 0);
			VME_Data_To_Mem(31 downto 0) <= X"0000" & VME_Data_In_lsw(15 downto 0);
			WR_mem <= '1';		 
				
			if (Transfer_Done = '1')	then 	NS <= end_cycle;
			else								NS <= set_lower_data; --wait
			end if;
			
	-------------------------------------------------------------------------------	
	--        WRITE MEMORY ADDRESS BITS TO LOWER FPGA, READ lsw FROM LOWER FPGA MEM								  
	-- 	Lower VME Read --		
			
		when get_lower_data =>	
			VME_Addr_To_Mem(25 downto 0) <= VME_Latched_Addr(25 downto 0);
			VME_Data_To_Mem(31 downto 0) <= X"00000000";-- does not matter what data is ...
			RD_mem <= '1';			
			
			if (Transfer_Done = '1')	then 	NS <= latch_lower_data; --read_lower_mem;
			else								NS <= get_lower_data; --wait
			end if;
			
		when latch_lower_data =>				NS <= end_cycle;
			Latch_VME_Dout <= '1';
		
	--------------------------------------------------------------------------------		
	-------------------------------------------------------------------------------	
	--        READ UPPER VME DATA BITS [31..16] FROM LOWER FPGA								  
	-- 	Upper VME Write --	
	--
	--  When writing the Upper fpga memory we must retreive the most significant
	--  bits of the VME data (31 ..16) from the Lower fpga.
	
		when get_upper_data =>					
			VME_Addr_To_Mem(25 downto 0) <= "00" & X"080000";
			VME_Data_To_Mem(31 downto 0) <= X"00000000"; -- does not matter what data is ...
			RD_mem <= '1';

			if (Transfer_Done = '1')	then 	NS <= latch_vme_data;
			else								NS <= get_upper_data; --wait
			end if;
			
		when latch_vme_data =>					NS <= setup_data_pause;
			Latch_VME_Din <= '1';
		
	--------------------------------------------------------------------------------		
	-------------------------------------------------------------------------------	
	--        WRITE UPPER VME DATA BITS [31..16] FROM UPPER MEM TO LOWER FPGA								  
	-- 	Upper VME Read --
					
		when set_upper_data =>				
			VME_Addr_To_Mem(25 downto 0) <= "00" & X"080000";
			VME_Data_To_Mem(31 downto 0) <= Latched_VME_Dout(31 downto 16) & X"0000";
								 
			WR_mem <= '1';
			if (Transfer_Done = '1')	then 	NS <= setup_data_pause;
			else								NS <= set_upper_data; --wait
			end if;
			
	-------------------------------------------------------------------------------	
		when setup_data_pause =>				NS <= end_cycle;
			VME_Addr_To_Mem(25 downto 0) <= VME_Latched_Addr(25 downto 0);
			VME_Data_To_Mem(31 downto 0) <= Latched_VME_Din(31 downto 0);
		
	-------------------------------------------------------------------------------	
	--                    COMPLETE THE VME DATA TRANSFER								  				
		when end_cycle =>					NS <= strobe_wr;
			VME_Addr_To_Mem(25 downto 0) <= VME_Latched_Addr(25 downto 0);
			VME_Data_To_Mem(31 downto 0) <= Latched_VME_Din(31 downto 0);
			
		when strobe_wr => 					NS <= dtacken;
			VME_Addr_To_Mem(25 downto 0) <= VME_Latched_Addr(25 downto 0);
			VME_Data_To_Mem(31 downto 0) <= Latched_VME_Din(31 downto 0);
			
			if(WR = '1') 	then 	VME_WR_Pulse <= '1';
			else						VME_WR_Pulse <= '0';
			end if;

		when dtacken => 
			DTACK <= '1';
			VME_Addr_To_Mem(25 downto 0) <= VME_Latched_Addr(25 downto 0);
			VME_Data_To_Mem(31 downto 0) <= Latched_VME_Din(31 downto 0);
			
			if (DS = '0')         	then  	NS <= xfer_done;
			else							NS <= dtacken; --wait
			end if;

		when xfer_done => 			
			if    (BS = '0')	  		then 	NS <= idle;
			elsif (HoldOff_VME = '1') 	then 	NS <= xfer_done;
			elsif (DS = '0')	  		then 	NS <= xfer_done;
			elsif (Block_xfer = '1') 	then 	NS <= data_setup_dly; 
				--go if block transfer and DS goes active while 
				--board is still selected and VME is not held off (BUSY).
			else					   			NS <= xfer_done;
			end if;

		when others =>              		NS <= idle;
	end case;
end process;


--===============================================================================
-- VME Address Decoding and Data Source/Destination Selection
-- Something to Realize !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
-- The address memory map given in the documentation for the crate processor 
-- programs includes an A[0] bit which is always assumed to be 0.                                   

ADDRESS_LATCH: dffr 
	generic map(26)
	port map(	d    => full_vme_addr,
				clk  => clk, 
				clrn => nreset, 
				ena  => LTCH_ADDR, 
				prn  => '1',
				q    => VME_Latched_Addr(25 downto 0)
			); 

VME_DATA_OUT_LATCH: dffr 
	generic map(16)
	port map(	d    => Mem_Data_To_VME(31 downto 16),
				clk  => clk, 
				clrn => nreset, 
				ena  => Latch_VME_Dout, 
				prn  => '1',
				q    => Latched_VME_Dout(31 downto 16)
			); 

VME_DATA_IN_LATCH: dffr 
	generic map(32)
	port map(	d    => Mem_Data_To_VME(31 downto 16) & VME_Data_In_lsw(15 downto 0),
				clk  => clk, 
				clrn => nreset, 
				ena  => Latch_VME_Din, 
				prn  => '1',
				q    => Latched_VME_Din(31 downto 0)
			); 


--===============================================================================
-- Outputs	

VME_RD_Lower_Mem <= RD_mem;
VME_WR_Lower_Mem <= WR_mem;

VME_Active_n <= NOT (Drive_VME AND ((NOT WRITE_n) OR (NOT DS0_n) OR (NOT DS1_n))); -- active low output
Write_nRead  <= WR; 
               
Upper_VME_Active_n <= VME_Active_n ; -- active low output
Upper_Write_nRead  <= Write_nRead;                
                    
Lower_VME_Active_n <= NOT( (NOT VME_Active_n) AND (NOT LWORD_n) );
Lower_Write_nRead  <= Write_nRead; 

VME_Write_Active <= WR AND BS;

BS_Out <= BS; -- diagnostic output
		
end;