Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 
USE work.DDS_VME_CONSTANTS.all;
-------------------------------------------------------------------------------
-- MODULE DESCRIPTION, March 2, 2011
--
-- The purpose of the Curve Interpolation process is to do the interpolation 
-- necessary to transform a set of [Time, Value] pairs (1023 pairs max) into 
-- as many as 10,240 curve points set at a constant time increment (typically 4us).

-- Version 6 differs from previous versions in that this module does not
-- poll the curve status words to see which curve needs to be interpolated.
-- This version is told which curve to work on through use of curve select
-- bits, curve_status[8..1].

-- The curve transformation is a simple linear interpolation of points at
-- defined times lying between two of the [Time, Value] pairs.  The two
-- pairs are denoted as [T(k), V(k)] and [T(k+1), V(k+1)].  We find the
-- interpolated curve value at each time t by:
--
--     C(t) = V(k) + ( t - T(k))*(V(k+1) - V(k)) / (T(k+1) - T(k))
--
-- The inputs to the computation, T(k+1) and V(k+1), are read from memory
-- at the begining and then again whenever time step t becomes greater than
-- T(k+1).  Note, [T(k),V(k)] is the previous value [T(k+1), V(k+1)] and 
-- are set by a register transfer.  The time step t is the output of a counter
-- internal to this process.  As a result, in order to produce a new output
-- from the interpolation computation, we only need to increment the time
-- step counter and perform two memory reads when t exceeds T(k+1).  There
-- are a maximum of 1023 [Time, Value] pairs that need to be read.  This is
-- typically a small fraction of the 10,240 points to be computed.
--
-- The other task that consumes time is writing the computation results to
-- the curve memory.
--
-- The curve interpolation computation takes approximately 26 clock cycles
-- to produce a result.  However, the computation is pipelined so that once
-- we have incurred the initial delay the computation will produce a result
-- as often as the input is updated.  Currently we are planning to update
-- the input every 6 clock cycles.  When the [Time, Value] pair must be updated
-- the "clock enable" for the computation blocks is taken away and the computation
-- halts while this is being done.  The computation is started again once all
-- of the inputs, including the time step, have been updated.  When the 
-- computation is halted the need to write results to curve memory pauses also.
-- This frees the databus to retreive T(k+1) and V(k+1) from memory. 
--
-- The logic implements a second small state machine that updates the time step
-- input variable and signals the main state machine when the [Time, Value] pair 
-- must be updated.  The small state machine, therefore manages the computation
-- inputs and strobes these into the computation.  The main state machine gets
-- into a small/fast loop looking for the computations output strobe and latching
-- results which are immediately written to memory.  When the small "input" state
-- machine signals that the [Time, Value] pair must be updated, the main state
-- machine disables the computation clock and branches off to update these inputs.
-- the main state machine signals the small state machine that the update is done
-- and then starts the computation and watches for output strobes again.

-- CURVE STATUS REGISTER: A curve status register of 8 bits is used to determine 
-- if a particular curve buffer is in need of interpolation.  The input 
-- "Preset_Status" is used to set all the bits of the status register when first 
-- booting up.  When a new curve buffer is loaded the status bit is set from the
-- Curve_Req_Process using the Set_Curve_Status bits.  Once a curve has been 
-- interpolated the status bit for that curve is reset.

-- ===========================================================================
-- Version 9  This version adds a counter to generate an address to write then
-- computation input variables and some intermediate variables to a test memory.
-- The variables are written with respect to the "Compute_Strobe".


entity Curve_Interp_Process_v9 is 
	port( 
		clk					: in std_logic;
		nreset				: in std_logic;
		Fkplus1_In	 		: in std_logic_vector(31 downto 0);
		UPD_Data_From_Mem : in std_logic_vector(31 downto 0);
		Transfer_Done		: in std_logic;
		Set_Curve_Status	: in std_logic_vector(8 downto 1);
		Preset_Status		: in std_logic;
		Curve_Intrp_Trigger	: in std_logic;

		Computation_Ready	: in std_logic;
		
		UPD_Data_to_Mem		: out std_logic_vector(31 downto 0);
		UPD_Addr_to_Mem		: out std_logic_vector(23 downto 0);
		UPD_WR_Pulse		: out std_logic;
		UPD_RD_Pulse		: out std_logic;
		UPD_Bus_Ena			: out std_logic;
		
		Run_Computation		: out std_logic;
		nReset_Computation	: out std_logic;
		Vkplus1				: out std_logic_vector(31 downto 0);
		Vk						: out std_logic_vector(31 downto 0);
		Tkplus1				: out std_logic_vector(31 downto 0);
		Tk						: out std_logic_vector(31 downto 0);
		T_Step				: out std_logic_vector(31 downto 0);		

		Compute_Strobe		: out std_logic;
		Crv_Intrp_Proc_Done	: out std_logic;  -- Strobes when SR is done
		
		TEST_MEMORY_PTR	: out std_logic_vector(7 downto 0)

		);
	end Curve_Interp_Process_v9;
	
architecture main of Curve_Interp_Process_v9 is	
-- Curve Status Register ----------------------------------------------
	signal curve_status		: std_logic_vector(8 downto 1);
	signal reset_status		: std_logic_vector(8 downto 1);
	
	signal status_to_clear	: std_logic_vector(8 downto 1);
	

-- Tigger inputs ------------------------------------------------------
	signal do_crv_interp	: std_logic;

-- VALUE_POINTER signals -------------------------------------------------
-- This counter acts as an address pointer to the "Value" of the [Time, Value]
-- pairs in the Shared Memory curve buffer.

	signal val_ptr_nrst		: std_logic; -- counter reset signal
	signal val_ptr_inc		: std_logic; -- counter increment enable strobe
	signal val_ptr_load		: std_logic;
	signal val_ptr_step		: std_logic_vector(15 downto 0);
	signal val_ptr_first		: std_logic_vector(19 downto 0);
	signal val_ptr_last		: std_logic_vector(19 downto 0);
--	signal val_ptr_done		: std_logic; -- count end flag
	signal val_ptr_out		: std_logic_vector(19 downto 0);
	
-- TIME_POINTER signals ------------------------------------------------
-- This counter acts as an address pointer to the "Time" of the [Time, Value]
-- pairs in the Shared Memory curve buffer.

	signal time_ptr_nrst		: std_logic; -- counter reset signal
	signal time_ptr_inc		: std_logic; -- counter increment enable strobe
	signal time_ptr_load		: std_logic;
	signal time_ptr_step		: std_logic_vector(15 downto 0);
	signal time_ptr_first	: std_logic_vector(19 downto 0);
	signal time_ptr_last		: std_logic_vector(19 downto 0);
	--signal time_ptr_done	: std_logic; -- count end flag
	signal time_ptr_out		: std_logic_vector(19 downto 0);

-- TIME_STEP_POINTER signals -------------------------------------------------
-- This counter increments and is used as the time step used in interpolating 
-- the final curve.  Typically this counter will step in increments of 4us.
	signal tstep_cnt_nrst	: std_logic; -- counter reset signal
	signal tstep_cnt_inc		: std_logic; -- counter increment enable strobe
	signal tstep_cnt_load	: std_logic;
	signal tstep_cnt_step	: std_logic_vector(3 downto 0);
	signal tstep_cnt_first	: std_logic_vector(31 downto 0);
	signal tstep_cnt_last	: std_logic_vector(31 downto 0);
	--signal tstep_cnt_done	: std_logic; -- count end flag
	signal tstep_cnt_out		: std_logic_vector(31 downto 0);

-- CURVE_POINTER signals -------------------------------------------------
-- This counter will act as the address pointer into the final curve memory.
-- It is used when writing the results of the interpolation computation.
	signal crv_ptr_nrst		: std_logic; -- counter reset signal
	signal crv_ptr_inc		: std_logic; -- counter increment enable strobe
	signal crv_ptr_load		: std_logic;
	signal crv_ptr_step		: std_logic_vector(3 downto 0);
	signal crv_ptr_first		: std_logic_vector(19 downto 0);
	signal crv_ptr_last		: std_logic_vector(19 downto 0);
	--signal crv_ptr_done	: std_logic; -- count end flag
	signal crv_ptr_out		: std_logic_vector(19 downto 0);

-- V(k) and V(k+1) Data Register -----------------------------------
	signal get_Vkplus1		: std_logic; 
	signal get_Vk				: std_logic;  
	signal Vk_out				: std_logic_vector(31 downto 0);
	signal Vkplus1_out		: std_logic_vector(31 downto 0);

-- T(k) and T(k+1) Data Register -----------------------------------
	signal get_Tkplus1		: std_logic; 
	signal get_Tk				: std_logic; 
	signal Tk_out				: std_logic_vector(31 downto 0);
	signal Tkplus1_In			: std_logic_vector(31 downto 0);
	signal Tkplus1_out		: std_logic_vector(31 downto 0);

-- F(k+1) Data Register (the computation output) -------------------
	signal get_Fkplus1		: std_logic;
	signal Fkplus1_out		: std_logic_vector(31 downto 0);

-- Memory Step Size Register Signals -------------------------------
-- Note: Some of the curves generated for the VME-DDS application are
-- 32 bit (4 byte) values and are addressed in 4 byte steps.  Other curves
-- are only 16 bit values and are addressed in 2 byte steps.  This
-- application provides for four 32 bit curves and four 16 bit curves.
-- The curves are arranged in memory so the four 16 bit curves follow
-- the four 32 bit curves.  The curve address pointer is tested, at one
-- point in the process and the required curve memory step is latched in
-- the memory_step register and then used as crv_ptr_step for the curve
-- address pointer.
	signal memory_step_d		: std_logic_vector(3 downto 0);
	signal get_memory_step	: std_logic;
	signal memory_step_q		: std_logic_vector(3 downto 0);

-- CURVE PROPERTIES ---------------------------------------------------------
-- The Properties word is defined as below:
--				Bit 31:  interpolated = 1, not interpolated = 0.
--				Bit 30:  32 bit data  = 1, 16 bit data = 0.
--				Bit 29:  read-only = 1, settable = 0.
--				Bit 28..20:  reserved
--				Bit 19..16:  time step, micro-sec (eg. 8us, 4us, 1us)
--				Bit 15..0 :  length, points (eg. 10,000 points, 1,000 points)
	signal crv_properties_d		: std_logic_vector(31 downto 0);
	signal get_crv_properties	: std_logic;
	signal crv_properties_q		: std_logic_vector(31 downto 0);

-- Table Size Counter ----------------------------------------------
-- This holds the "Table Size" value that was set as the first word
-- in the curve buffer.  This value is needed to determine when the
-- list of "Values" end and the list of "Times" begin.  It is also
-- needed to make the process aware of when the user provided 
-- [Time, Value] pairs has been exhausted.
	signal table_cnt_nrst	: std_logic; -- counter reset signal
	signal table_cnt_dec	: std_logic; -- counter increment enable strobe
	signal table_cnt_load	: std_logic;
	signal table_cnt_first	: std_logic_vector(10 downto 0);
	signal table_cnt_done	: std_logic; -- count end flag
	signal table_cnt_out		: std_logic_vector(10 downto 0);

-- END COMPUTATION COUNTER ----------------------------------------------
-- This counter is triggered when the table_cnt counter indicates that we
-- have used all of the available Time,Value pairs.  We do still have 
-- computations in the pipeline so we want to count the Computation_ready
-- pulses to decide when the pipline has been emptied of valid results.
	signal end_count				: std_logic_vector(3 downto 0);
	signal computation_ended	: std_logic;

-- Transfer Word Counter -------------------------------------------
-- This counter counts the number of writes to the final curve memory.
-- This will tell us when we have written the full number of curve points.
	signal xfer_nrst		: std_logic;		
	signal xfer_inc		: std_logic;	
	signal xfer_done		: std_logic;
	
-- Pause Timer -----------------------------------------------------
	signal pause_tmr 		: std_logic_vector(2 downto 0);
	signal inc_pause_tmr	: std_logic;
  
-- Pause Timer2 -----------------------------------------------------
	signal pause_tmr2 	 : std_logic_vector(2 downto 0);
	signal inc_pause_tmr2	: std_logic;

-- Transfer Done Latch ---------------------------------------------
	signal transf_done		: std_logic;
	signal clear_transf_done: std_logic;

-- Inter-State Machine Signals -------------------------------------
	signal  	run				: std_logic;
	signal	start_input  	: std_logic;
	signal  	intrp_done		: std_logic;
	signal  	do_update 		: std_logic;
	signal	update_done	 	: std_logic;
	
-- Test Memory Address Pointer -------------------------------------
	signal	test_mem_rst	: std_logic;
	signal	test_mem_inc	: std_logic;
	signal	test_mem_ptr	: std_logic_vector(7 downto 0);
	
-- STATE_MACHINE ---------------------------------------------------
	type STATE_TYPE is
(
	idle, init_crv_proc, test_crv_valid, setup_tbl_cnt, setup_val_ptr,
	setup_time_ptr1, setup_time_ptr2, chk_intrp_bit, transfer_crv1, 
	transfer_crv2, transfer_crv3, transfer_crv4, transfer_crv5, 
	init_Vk, inc_val_ptr_0, init_Vkplus1, inc_val_ptr_1, init_Tk, inc_time_ptr_0, 
	init_Tkplus1, inc_time_ptr_1, start_computation, wait_for_output, 
	wr_curve_memory, inc_crv_ptr, wait_for_transf_done, update_Vk_Tk, 
	update_Vkplus1, update_Tkplus1, inc_vt_ptrs, done, finish
);		
 	signal CS, NS : STATE_TYPE;	

-- STATE_MACHINE 2---------------------------------------------------
	type STATE_TYPE2 is
(
	idle2, test_time_step, strobe_input, pause, inc_test_mem_ptr, 
	inc_time_step, signal_update
);		
 	signal CS2, NS2 : STATE_TYPE2;	

--########################################################
--========================================================
--						B E G I N
--========================================================
--########################################################

begin

--**-- --**--   --**--   --**--   --**--   --**--   --**-- 
-- STROBE SYNCHRONIZATION --------------------------------  
SYNC_UPDATE: led -- syncronize the signal to start update polling sequence.
port map(	sig_in  => Curve_Intrp_Trigger, 
				clk     => clk, 
				clrn    => '1', 
				pls_out => do_crv_interp); 

	
-- VALUE_POINTER_COUNTER --------------------------------------------
VALUE_ADDR_POINTER: countby_n -- Address Pointer into Curve Buffer
generic map(20,16)
port map(	clk  	 => clk,
				clrn 	 => val_ptr_nrst,
				ena  	 => val_ptr_inc,
				load 	 => val_ptr_load,
				ld_data  => val_ptr_first,
				step	 	=> val_ptr_step,
				flag_cnt => val_ptr_last,
				flag_out => open, --val_ptr_done,
				cnt_out  => val_ptr_out 
		);

-- TIME_POINTER_COUNTER -------------------------------------------
TIME_ADDR_POINTER: countby_n -- Address Pointer into Curve Buffer
generic map(20,16)
port map(	clk  	 => clk,
			clrn 	 => time_ptr_nrst,
			ena  	 => time_ptr_inc,
			load 	 => time_ptr_load,
			step	 => time_ptr_step,
			ld_data  => time_ptr_first,
			flag_cnt => time_ptr_last,
			flag_out => open, --time_ptr_done,
			cnt_out  => time_ptr_out 
		);

-- CURVE_ADDR_POINTER_COUNTER ---------------------------------------
CURVE_ADDR_POINTER: countby_n -- Address Pointer to Curve Memory
generic map(20,4)
port map(	clk  	 => clk,
			clrn 	 => crv_ptr_nrst,
			ena  	 => crv_ptr_inc,
			load 	 => crv_ptr_load,
			ld_data  => crv_ptr_first,
			step	 => crv_ptr_step,
			flag_cnt => X"FFFFF",
			flag_out => open, --crv_ptr_done,
			cnt_out  => crv_ptr_out 
		);

-- TRANSFER_COUNTER ------------------------------------------------
TRANSFER_COUNT: counter -- Counts words transfered
generic map(16)
port map(	clk  	 => clk,
			clrn 	 => xfer_nrst,
			ena  	 => xfer_inc,
			load 	 => '0',
			updn	 => '1',
			ld_data  => (others => '0'),
			flag_cnt => crv_properties_q(15 downto 0), --"0010011100010000", -- 10,000
			flag_out => xfer_done,
			cnt_out  => open 
		);

-- V(k+1) DATA REGISTER -----------------------------------------------------
Vkplus1_DATA_REGISTER: dffr
generic map(32)
port map( 	d    => UPD_Data_From_Mem, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_Vkplus1, 
			q    => Vkplus1_out 
		);

	Vkplus1 <= Vkplus1_out;

-- V(k) DATA REGISTER -------------------------------------------------------
Vk_DATA_REGISTER: dffr
generic map(32)
port map( 	d    => Vkplus1_out, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_Vk, 
			q    => Vk_out 
		);
		
	Vk <= Vk_out;
	
-----------------------------------------------------------------------------
-- T(k+1) DATA REGISTER ------------------------------------------------
Tkplus1_DATA_REGISTER: dffr
generic map(32)
port map( 	d    => Tkplus1_In, --is actually UPD_Data_From_Mem, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_Tkplus1, 
			q    => Tkplus1_out 
		);

	Tkplus1 <= Tkplus1_out;  

-- T(k) DATA REGISTER ------------------------------------------------
Tk_DATA_REGISTER: dffr
generic map(32)
port map( 	d    => Tkplus1_out, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_Tk,
			q    => Tk_out 
		);
		
	Tk <= Tk_out;
	
-- F(k+1) DATA REGISTER -----------------------------------------------
Fkplus1_DATA_REGISTER: dffr
generic map(32)
port map( 	d    => Fkplus1_In, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_Fkplus1, -- Computation_Ready AND NOT computation_ended;
			q    => Fkplus1_out 
		);
		
		get_Fkplus1 <= Computation_Ready AND NOT computation_ended; 
		
-- MEMORY STEP REGISTER -----------------------------------------------
MEM_STEP_REGISTER: dffr
generic map(4)
port map( 	d    => memory_step_d, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_memory_step, 
			q    => memory_step_q 
		);
		
-- CURVE PROPERTIES REGISTER ------------------------------------------
CURVE_PROPERTIES_REGISTER: dffr
generic map(32)
port map( 	d    => crv_properties_d, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_crv_properties, 
			q    => crv_properties_q 
		);

--TABLE SIZE COUNTER ------------------------------------------------
TABLE_SIZE_COUNTER: counter -- COUNTS DOWN AVAILABLE [TIME, VALUE] PAIRS
generic map(11)
port map(	clk  	 => clk,
				clrn 	 => table_cnt_nrst,
				ena  	 => table_cnt_dec,
				updn	 => '0', -- count down
				load 	 => table_cnt_load,
				ld_data  => table_cnt_first,
				flag_cnt => "00000000000",  --counts down to zero
				flag_out => table_cnt_done,
				cnt_out  => table_cnt_out      
		);

-- TEST MEMORY ADDRESS POINTER -------------------------------------------
-- Added in version 9 to provide an address to the test memory recording
-- input and intermediate variables used in the computation.
TEST_MEM_ADDRESS_PTR:
process (clk, nreset, test_mem_rst, test_mem_inc, test_mem_ptr)
begin
	if (nreset = '0') then test_mem_ptr <= (others => '0');
	elsif rising_edge(clk) then		
		if   (test_mem_rst = '1') then 	test_mem_ptr <= (others => '0');
		elsif(test_mem_ptr = X"FF") then	test_mem_ptr <= test_mem_ptr;	
		elsif(test_mem_inc = '1') then 	test_mem_ptr <= test_mem_ptr + 1;
		else										test_mem_ptr <= test_mem_ptr;												
		end if; 
	end if;
end process;

					TEST_MEMORY_PTR <= test_mem_ptr;

--END COMPUTATION COUNTER ------------------------------------------------
-- Once the last Time Value pair has been submitted to the pipelined
-- interpolation computation and time steps past the last Time point
-- the pipeline needs a number of clocks to empty the pipline.  

END_COMPUTATION_COUNTER: 
process
(clk, table_cnt_done, end_count, Computation_Ready)
begin
	if rising_edge(clk) then		
		if   (table_cnt_done = '0') then end_count <= (others => '0');
		elsif(end_count < "0100") AND (Computation_Ready = '1') then 
												end_count <= end_count + 1;
		end if; 
		
		if   (end_count = "0100") then computation_ended <= '1';
		else                           computation_ended <= '0';
		end if;
	end if;
end process;
			
-----------------------------------------------------------------------------
-- TIME_STEP_COUNTER --------------------------------------------
TIME_STEP_COUNTER: countby_n -- Produces a time step value for computation
generic map(32,4)
port map(	clk  	 => clk,
				clrn 	 => tstep_cnt_nrst,
				ena  	 => tstep_cnt_inc,
				load 	 => tstep_cnt_load,
				ld_data  => tstep_cnt_first, -- <= (others => '0');
				step	 	=> tstep_cnt_step,  -- <= crv_properties_q(19 downto 16);
				flag_cnt => tstep_cnt_last,  -- <= (others => '1');
				flag_out => open, 			  -- Not Considered
				cnt_out  => tstep_cnt_out 
		);
		
	T_Step <= tstep_cnt_out;

------------------------------------------------------------------------------		
-- CURVE SELECT LOGIC --------------------------------------------------------
CURVE_STATUS_REGISTER:
process (clk, nreset, Preset_Status, reset_status)
begin
	if   (nreset = '0') then 
		curve_status <= (others => '0');
	elsif rising_edge(clk) then
		if  (Preset_Status = '1') then
			curve_status <= (others => '1');
		else
			curve_status <= (curve_status OR Set_Curve_Status) AND not reset_status;
		end if;
	end if;
end process;

curve_status_LOGIC:
process(curve_status(8 downto 1))
begin
	if   (curve_status(1) = '1') then
		crv_ptr_first 	 <= FREQ_CURVE_ADDR(19 downto 0); --Address of RAM curve (constant)
		crv_properties_d<= FREQ_CURVE_PROP; --The curve property word(constant)
		val_ptr_first	 <= SM_CURVE_TBL1_SIZE(19 downto 0); -- address (# of pairs)
		status_to_clear <= "00000001";
		
	elsif(curve_status(2) = '1') then	
		crv_ptr_first 	 <= BIAS_CURVE_ADDR(19 downto 0); --Address of RAM curve (constant)
		crv_properties_d<= BIAS_CURVE_PROP; --The curve property word(constant)
		val_ptr_first	 <= SM_CURVE_TBL2_SIZE(19 downto 0); -- address (# of pairs)
		status_to_clear <= "00000010";
		
	elsif(curve_status(3) = '1') then	
		crv_ptr_first 	<= INJECT_CURVE_ADDR(19 downto 0); --Address of RAM curve (constant)
		crv_properties_d<= X"00000000"; --The curve property word(constant) 
													-- Not interpolated, Not loaded from TABLE 3
													-- Injection Cuve memory load with Frequency Curve.
													-- was before 3/24/15 INJECT_CURVE_PROP; 
		val_ptr_first	<= SM_CURVE_TBL3_SIZE(19 downto 0); -- address (# of pairs)
		status_to_clear <= "00000100";
		
	elsif(curve_status(4) = '1') then	
		crv_ptr_first 	<= SPARE_32B_ADDR(19 downto 0); --Address of RAM curve (constant)
		crv_properties_d<= SPARE_32B_PROP; --The curve property word(constant)
		val_ptr_first	<= SM_CURVE_TBL4_SIZE(19 downto 0); -- address (# of pairs)
		status_to_clear <= "00001000";
		
	elsif(curve_status(5) = '1') then	
		crv_ptr_first 	<= POFF1_CURVE_ADDR(19 downto 0); --Address of RAM curve (constant)
		crv_properties_d<= POFF1_CURVE_PROP; --The curve property word(constant)
		val_ptr_first	<= SM_CURVE_TBL5_SIZE(19 downto 0); -- address (# of pairs)
		status_to_clear <= "00010000";
		
	elsif(curve_status(6) = '1') then	
		crv_ptr_first 	<= POFF2_CURVE_ADDR(19 downto 0); --Address of RAM curve (constant)
		crv_properties_d<= POFF2_CURVE_PROP; --The curve property word(constant)
		val_ptr_first	<= SM_CURVE_TBL6_SIZE(19 downto 0); -- address (# of pairs)
		status_to_clear <= "00100000";
		
	elsif(curve_status(7) = '1') then	
		crv_ptr_first 	<= PARAPH_CURVE_ADDR(19 downto 0); --Address of RAM curve (constant)
		crv_properties_d<= PARAPH_CURVE_PROP; --The curve property word(constant)
		val_ptr_first	<= SM_CURVE_TBL7_SIZE(19 downto 0); -- address (# of pairs)
		status_to_clear <= "01000000";
		
	elsif(curve_status(8) = '1') then	
		crv_ptr_first 	<= SPARE_16B_ADDR(19 downto 0); --Address of RAM curve (constant)
		crv_properties_d<= SPARE_16B_PROP; --The curve property word(constant)
		val_ptr_first	<= SM_CURVE_TBL8_SIZE(19 downto 0); -- address (# of pairs)
		status_to_clear <= "10000000";
		
	else	
		crv_ptr_first 	<= FREQ_CURVE_ADDR(19 downto 0); --Address of RAM curve (constant)
		crv_properties_d<= X"00000000"; --The curve property word(constant)
		val_ptr_first	<= SM_CURVE_TBL1_SIZE(19 downto 0); -- address (# of pairs)
		status_to_clear <= "00000000";
	end if;
end process;
------------------------------------------------------------------------------
------------------------------------------------------------------------------		
PAUSE_TIMER: -- Timer used to generate setup time for FLASH writes
process (clk, inc_pause_tmr)
begin
	if rising_edge(clk) then
		if   (inc_pause_tmr = '0') then pause_tmr <= "000";
		elsif(inc_pause_tmr = '1') then pause_tmr <= pause_tmr + 1;
		end if;
	end if;
end process;
	
------------------------------------------------------------------------------		
PAUSE_TIMER2: -- Timer used to generate setup time for FLASH writes
process (clk, inc_pause_tmr2)
begin
	if rising_edge(clk) then
		if   (inc_pause_tmr2 = '0') then pause_tmr2 <= "000";
		elsif(inc_pause_tmr2 = '1') then pause_tmr2 <= pause_tmr2 + 1;
		end if;
	end if;
end process;

------------------------------------------------------------------------------
TRANSFER_DONE_LATCH: 
process (clk, nreset, Transfer_Done, clear_transf_done)
begin
	if(nreset = '0')       				then transf_done <= '0';
	elsif rising_edge(clk) then
		if   (clear_transf_done = '1') 	then transf_done <= '0';
		elsif(Transfer_Done = '1')     	then transf_done <= '1';
		end if;
	end if;
end process;
	
--**--   --**--   --**--   --**--   --**--   --**--   --**--  
----------------------------------------------------
-- STATE MACHINE  
----------------------------------------------------
SM_REG: process(clk, nreset)
begin
	if(nreset='0')				then CS <= idle;
	elsif(rising_edge(clk))		then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process
(CS, do_crv_interp, val_ptr_out, pause_tmr,
 Vkplus1_out, time_ptr_out, UPD_Data_From_Mem, table_cnt_out, xfer_done,
 tstep_cnt_out, Tkplus1_out, table_cnt_done, crv_ptr_out, Fkplus1_out,
 memory_step_q, Computation_Ready, do_update, transf_done, crv_properties_q,
 status_to_clear)
begin
	-- Defaults -------------------------------------
		UPD_Addr_to_Mem <= (others => '0');
		UPD_Data_to_Mem <= (others => '0');
		UPD_WR_Pulse 	<= '0';
		UPD_RD_Pulse 	<= '0';
		UPD_Bus_Ena		<= '1';  -- enabled except in the idle state
		
		val_ptr_nrst	<= '1';
		val_ptr_inc		<= '0';
		val_ptr_load	<= '0';
		val_ptr_step	<= X"0004";
--		val_ptr_first	<= (others => '0'); set by curve_status
		val_ptr_last	<= SM_CURVE_TBL8_STAT(19 downto 0);   -- !@#$!@#$ suspect if using table 8
		
		time_ptr_nrst 	<= '1';
      time_ptr_inc 	<= '0'; 
      time_ptr_load 	<= '0';
		time_ptr_step	<= X"0004";
      time_ptr_first	<= (others => '0');
		time_ptr_last	<= (others => '0');
		
		tstep_cnt_nrst	<= '1';
		--tstep_cnt_inc	<= '0'; -- incremented in State Machine 2
		tstep_cnt_load	<= '0';
		tstep_cnt_step	<= crv_properties_q(19 downto 16);
		tstep_cnt_first<= (others => '0');
		tstep_cnt_last	<= (others => '1');

		crv_ptr_nrst	<= '1';
		crv_ptr_inc		<= '0';
		crv_ptr_load	<= '0';
		crv_ptr_step	<= memory_step_q;

		table_cnt_nrst	<= '1';
		table_cnt_dec	<= '0';
		table_cnt_load	<= '0';
		table_cnt_first	<= (others => '0');
		
		get_Vkplus1		<= '0';  -- Strobe to latch data into register
		get_Tkplus1 	<= '0';
		get_Vk			<= '0';
		get_Tk			<= '0';
		
		Tkplus1_In  	<= UPD_Data_From_Mem;
		
		get_crv_properties 	<= '0';
		
		memory_step_d 		<= X"0";
		get_memory_step 	<= '0';			
		
		xfer_nrst			<= '1';	
		xfer_inc 			<= '0';	
                
		inc_pause_tmr		<= '0';

		start_input			<= '0';
		update_done			<= '0';
		intrp_done			<= '0';
				
		run 					<= '0';
		clear_transf_done <= '0';
		
		reset_status		<= (others => '0');

		nReset_Computation 	<= '1';
		Crv_Intrp_Proc_Done <= '0';
		
		test_mem_rst <= '0';
	--------------------------------------------------

	case CS is
		when idle =>
			UPD_Bus_Ena		<= '0'; -- enabled except in the idle state
			
			nReset_Computation 	<= '0';
			val_ptr_nrst		<= '0';
			time_ptr_nrst		<= '0';
			tstep_cnt_nrst		<= '0';
			crv_ptr_nrst		<= '0';
			table_cnt_nrst		<= '0';
			xfer_nrst			<= '0';	
			intrp_done			<= '1';

			test_mem_rst 		<= '1';
		
			if   (do_crv_interp = '1')
			then 									NS <= init_crv_proc;
			else                   			NS <= idle;
			end if;

		----------------------------------------------------------------
		-- Initialize registers according to the curve_status bits in.
		
		when init_crv_proc => 				NS <= test_crv_valid;
		-- The val_ptr counter will double as the pointer to the status
		-- words when looking for the curve that needs to be updated.
			val_ptr_load			<= '1'; --val_ptr_first	<= SM_CURVE_TBLx_SIZE
			crv_ptr_load			<= '1';
			get_crv_properties	<= '1'; 
			--^--if nothing needs done, crv_properties_q <= X"00000000"
		
		----------------------------------------------------------------
		-- Test properties to see if the curve is defined, count > 0
		when test_crv_valid => 
			if   (crv_properties_q(15 downto 0) = X"0000") then	
											NS <= done;
			else 							NS <= setup_tbl_cnt;
			end if;
		
		----------------------------------------------------------------
		-- Recall that "table_cnt" is a count down counter.		
		when setup_tbl_cnt => 
			table_cnt_load	<= '1';
			UPD_Addr_to_Mem <= "0000" & val_ptr_out(19 downto 0); -- = SM_CURVE_TBLx_SIZE
			table_cnt_first	<= UPD_Data_From_Mem(10 downto 0); -- Number of time,value pairs
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (pause_tmr = "011") then	NS <= setup_val_ptr;
			else									NS <= setup_tbl_cnt;
			end if;
					
		----------------------------------------------------------------
		-- Here we increment the "val_ptr" so that we are pointing to the
		-- first value.
		when setup_val_ptr =>					NS <= setup_time_ptr1;
			UPD_Addr_to_Mem <= "0000" & val_ptr_out(19 downto 0); --held over
			val_ptr_inc 	<= '1'; -- step to first value point word
			val_ptr_step	<= X"0004";	 
		
		----------------------------------------------------------------				
		when setup_time_ptr1 =>					NS <= setup_time_ptr2;
				-- set time pointers equal to the value pointer
			time_ptr_load	<= '1';
			time_ptr_first	<= val_ptr_out; 
			
				-- determine if we have a 16 bit or 32 bit curve to update
				-- and latch the step size we will be using.
			get_memory_step <= '1';			
			if   (crv_properties_q(30) = '0') then
				-- The curve to be updated is a 16 Bit Curve
				memory_step_d <= X"2";
			else
				-- The curve to be updated is a 32 Bit Curve
				memory_step_d <= X"4";
			end if;
		
		----------------------------------------------------------------
		-- Recall that the Curve Buffers are written with all the values
		-- followed by all the associated times.  Time ptr step below steps
		-- the time_ptr from the first "value" to the first "time".
		
		when setup_time_ptr2 =>					NS <= chk_intrp_bit;
				time_ptr_inc 	<= '1'; -- step to curves first time point
				time_ptr_step	<= "000" & table_cnt_out(10 downto 0) & "00";	 
								-- note table_cnt_out is multiplied by 4
								-- (4 bytes per 32 bit word).
				tstep_cnt_nrst	<= '0'; -- start curve time step at zero
		
		----------------------------------------------------------------
		-- At this point we have:
		--   1. val_ptr pointing at the fist value V(k+1)
		--   2. time_ptr pointing at the first time T(k+1)
		--   3. the interpolated curve time step is at zero.
		--   4. the crv_ptr is pointing at the first curve memory word 
		--      to be written.
		--   5. the curve memory step size has been set.
		--   6. the table size has been latched.
		--
		--  Now, let's begin the transformation ...
		--
		-- First check that the curve needs to be interpolated or simply
		-- transferred as is.
		
		when chk_intrp_bit =>
			if(crv_properties_q(31) ='1') then 	NS <= init_Vk;
			else											NS <= transfer_crv1;
			end if;
		
		when transfer_crv1 =>
			get_Vkplus1	<= '1';
			UPD_Addr_to_Mem <= "0000" &  val_ptr_out(19 downto 0);
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (pause_tmr = "011") then	NS <= transfer_crv2;
			else									NS <= transfer_crv1;
			end if;
		
		when transfer_crv2 => 				NS <= transfer_crv3;
			val_ptr_inc 	<= '1'; 
			val_ptr_step	<= X"0004";
			
			clear_transf_done <= '1'; -- this process latches the InterFPGA
									  -- Transfer Done signal so we must be 
									  -- sure to clear it before we use it.

		when transfer_crv3 => 				
			UPD_Addr_to_Mem <= "0000" & crv_ptr_out(19 downto 0);
			UPD_Data_to_Mem <= Vkplus1_out;     
			UPD_WR_Pulse 	<= '1'; -- << The write operation
			
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (crv_ptr_out < X"20000") then
				if  (pause_tmr = "011") then	NS <= transfer_crv4;
				else									NS <= transfer_crv3;
				end if;
			elsif(transf_done = '1') then NS <= transfer_crv4;  -- can we get a rogue transf_done ??
			else									NS <= transfer_crv3;
			end if;

		when transfer_crv4 => 				NS <= transfer_crv5;
			crv_ptr_inc 	<= '1'; 
			crv_ptr_step	<= memory_step_q;
			
			table_cnt_dec	<= '1'; 	-- Count to number of points in curve buffer
											-- This counter will stop decrementing at zero.
			xfer_inc 		<= '1';	-- increment the count of curve writes made
											-- Count to end of the curve memory
		when transfer_crv5 => 				
			if   (xfer_done = '1') 		then	NS <= done;
			elsif(table_cnt_done = '1') then	NS <= transfer_crv2; --done;
				-- going to transfer_crv2 will avoid changing the final value.
			else										NS <= transfer_crv1;
			end if;
		
		----------------------------------------------------------------
		--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
		-- initialize input variables
		--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
		----------------------------------------------------------------
		-- Set up the initial values of V(k), V(k+1), T(k) and T(k+1) into
		-- the computation input variable registers.
		when init_Vk =>						
			get_Vkplus1	<= '1';     -- This will be copied to Vk next
			UPD_Addr_to_Mem <= "0000" & val_ptr_out(19 downto 0);
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (pause_tmr = "011") then	NS <= inc_val_ptr_0;
			else									NS <= init_Vk;
			end if;
				
		when inc_val_ptr_0 =>				NS <= init_Vkplus1;
			get_Vk	<= '1';			-- Vk_out <= Vkplus1_out
			val_ptr_inc 	<= '1'; 
			val_ptr_step	<= X"0004";	 

		when init_Vkplus1 =>				
			get_Vkplus1	<= '1';
			UPD_Addr_to_Mem <= "0000" & val_ptr_out(19 downto 0);
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (pause_tmr = "011") then	NS <= inc_val_ptr_1;
			else									NS <= init_Vkplus1;
			end if;		
		
		when inc_val_ptr_1 =>				NS <= init_Tk;
			val_ptr_inc 	<= '1'; 
			val_ptr_step	<= X"0004";	 
		
		when init_Tk =>						
			get_Tkplus1	<= '1';
			UPD_Addr_to_Mem <= "0000" & time_ptr_out(19 downto 0);
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (pause_tmr = "011") then	NS <= inc_time_ptr_0;
			else									NS <= init_Tk;
			end if;

		when inc_time_ptr_0 =>				NS <= init_Tkplus1;
			get_Tk	<= '1';
			time_ptr_inc 	<= '1'; 
			time_ptr_step	<= X"0004";	 
			
			--table_cnt_dec	<= '1'; --!!@#$!@#$%#$%^$#%^$%^&&^%$&^@$#%!@#$!~@#~!@#@$#%^

		when init_Tkplus1 =>				
			get_Tkplus1	<= '1';
			UPD_Addr_to_Mem <= "0000" & time_ptr_out(19 downto 0);
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (pause_tmr = "011") then	NS <= inc_time_ptr_1;
			else									NS <= init_Tkplus1;
			end if;
		
		when inc_time_ptr_1 =>				NS <= start_computation;
			time_ptr_inc 	<= '1'; 
			time_ptr_step	<= X"0004";	 
			
			table_cnt_dec	<= '1';
			
			run <= '1'; -- :::::: 
		
		----------------------------------------------------------------
		--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
		-- start computation loop
		--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
		----------------------------------------------------------------
		when start_computation =>			NS <= wait_for_output;
			run 	<= '1'; -- :::::: 
			start_input			<= '1';
		
		when wait_for_output =>
			run 	<= '1'; -- :::::: 
			clear_transf_done <= '1';
			
			if   (xfer_done = '1')		   		then NS <= done;
			elsif(Computation_Ready = '1') 		then NS <= wr_curve_memory;	
			elsif(do_update = '1')   	   		then NS <= update_Vk_Tk;
			else									 			  NS <= wait_for_output;
			end if;
				
		when wr_curve_memory =>						
			run <= '1'; -- :::::: 
			UPD_Addr_to_Mem <= "0000" & crv_ptr_out(19 downto 0);
			UPD_Data_to_Mem <= Fkplus1_out;
			UPD_WR_Pulse 	<= '1'; -- << The write operation
			
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "011") then		
				if   (crv_ptr_out < X"20000") then 	NS <= inc_crv_ptr; -- curve in upper fpga
				elsif(transf_done = '1')	  then 	NS <= inc_crv_ptr; -- curve in lower fpga
				else											NS <= wait_for_transf_done;
				end if;
			else												NS <= wr_curve_memory;
			end if;
		
		when wait_for_transf_done => 		
			run <= '0'; -- ##### computations get held up waiting for interfpga transfers.
			UPD_Addr_to_Mem <= "0000" & crv_ptr_out(19 downto 0);
			UPD_Data_to_Mem <= Fkplus1_out;
			UPD_WR_Pulse 	<= '1'; -- << The write operation
			
			if(transf_done = '1')			  then 	NS <= inc_crv_ptr;
			else												NS <= wait_for_transf_done;
			end if;
			
		when inc_crv_ptr =>								NS <= wait_for_output;
			run <= '1'; -- :::::: 
			
			crv_ptr_inc 	<= '1'; 
			crv_ptr_step	<= memory_step_q;
						
			xfer_inc 		<= '1';	-- increment the count of computations made
			
		----------------------------------------------------------------
		--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
		-- Vk and Tk Update Block
		--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
		----------------------------------------------------------------
		when update_Vk_Tk =>				NS <= update_Vkplus1;
			--run <= '0'; -- by default
			get_Vk			<= '1';	-- update V(k) <- V(k+1)
			get_Tk			<= '1';	-- update T(k) <- T(k+1)
				
		when update_Vkplus1 =>	
			get_Vkplus1		<= '1';
			UPD_Addr_to_Mem <= "0000" & val_ptr_out(19 downto 0);
		
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (pause_tmr = "011") then	NS <= update_Tkplus1;
			else									NS <= update_Vkplus1;
			end if;		
		
		when update_Tkplus1 =>					
			get_Tkplus1		<= '1';
			UPD_Addr_to_Mem <= "0000" & time_ptr_out(19 downto 0);
					
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (pause_tmr = "111") then	NS <= inc_vt_ptrs;
			else									NS <= update_Tkplus1;
			end if;

		when inc_vt_ptrs =>					NS <= wait_for_output;		
			val_ptr_inc 	<= '1'; 
			val_ptr_step	<= X"0004";	
			
			time_ptr_inc 	<= '1'; 
			time_ptr_step	<= X"0004";	 
			
			table_cnt_dec	<= '1';
			update_done		<= '1';

		----------------------------------------------------------------
		
		when done =>						NS <= idle;
			Crv_Intrp_Proc_Done 	<= '1';
			intrp_done				<= '1';
			reset_status			<= status_to_clear;
			
		when others =>              		NS <= idle;
	end case;
end process;

--**--   --**--   --**--   --**--   --**--   --**--   --**--  
----------------------------------------------------
-- STATE MACHINE2  
----------------------------------------------------
SM_REG2: process(clk, nreset, intrp_done, run)
begin
	if(nreset='0') OR (intrp_done = '1') then CS2 <= idle2;
	elsif(rising_edge(clk))				 then CS2 <= NS2;
	end if;
end process;

SM_NEXT_STATE2: 
process
(CS2, start_input, update_done, intrp_done, Tkplus1_out, tstep_cnt_out,
 table_cnt_done, pause_tmr2, run)
begin
	-- Defaults -------------------------------------
		tstep_cnt_inc	<= '0';
		inc_pause_tmr2	<= '0';
		
		do_update 		<= '0';
						
		Compute_Strobe <= '0';	
		test_mem_inc   <= '0';

	--------------------------------------------------

	case CS2 is
	
		----------------------------------------------------------------
		--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
		-- start computation loop  
		-- Tests when 't' > Tkplus1 then signals update of Tk, Tkplus1, Vk, Vkplus1
		-- Sends Compute strobe every 7 clocks.
		-- increments 't' 6 clocks after Compute strobe.
		--::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::--
		----------------------------------------------------------------
		when idle2 =>						
			if(start_input = '1') then		NS2 <= test_time_step;
			else									NS2 <= idle2;
			end if;
		
		when test_time_step =>			
			if   (intrp_done = '1')			then  NS2 <= idle2; --!@#$@#!$@!  Added 3/23/15
			elsif(run = '0') 					then	NS2 <= test_time_step;
			elsif(Tkplus1_out < tstep_cnt_out)
				AND (table_cnt_done = '0') then 	NS2 <= signal_update;
			else											NS2 <= strobe_input;
			end if;
					
		when strobe_input =>
			
			if   (run = '0')	then				NS2 <= strobe_input;
					Compute_Strobe 	<= '0'; -- <-- the marker
			else										NS2 <= pause;
					Compute_Strobe 	<= '1'; -- <-- the marker
			end if;
							
		when pause =>	-- spaces the strobes enough to manage the computation output				
			
			if   (run = '0')		    then		NS2 <= pause;
				inc_pause_tmr2 <= '0';
			else
				-- increment the pause timer and wait for timeout
				inc_pause_tmr2	<= '1'; 			
				if (pause_tmr2 = "010") then	NS2 <= inc_test_mem_ptr; -- pause was reduced by 1
				else									NS2 <= pause;				 -- when inc_test_mem_ptr was added
				end if;
			end if;
		
		when inc_test_mem_ptr =>				NS2 <= inc_time_step;
			test_mem_inc   <= '1';
		
		when inc_time_step =>				
		
			if   (run = '0')			then		NS2 <= inc_time_step;
				tstep_cnt_inc	<= '0'; 
			else										NS2 <= test_time_step;
				tstep_cnt_inc	<= '1'; -- advance the time step input variable
			end if;
	
		----------------------------------------------------------------------	
		when signal_update =>
			do_update 		<= '1';
			
			if(update_done = '1') then 		NS2 <= test_time_step;
			else										NS2 <= signal_update;
			end if;
		
		when others =>              			NS2 <= idle2;
	end case;
end process;

	Run_Computation <= run;
	
end;