Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 
USE work.DDS_VME_CONSTANTS.all;

-------------------------------------------------------------------------------
-- DDS_App_Sequence
--
-- This is the main sequence of events process.  It performs the setup of all
-- the shared memory from Flash memory at power up and then updates all the 
-- operational parameter registers, and updates and interpolates all of the
-- DDS frequency and phase curves from the shared memory into the curve
-- RAM memory.
--
-- Once everything is restored and initialized this process triggers other processes
-- to check for parameter updates in the shared memory and to check for requests
-- for curve buffer data or updates to curve buffer data. It will also poll the
-- status of the curve buffers to determine when they have been updated and the
-- associated DDS curve RAM memory needs to be updated from its buffer.
--
-------------------------------------------------------------------------------
-- Power Up Sequence:
-- 1. Restore the Shared Memory from the Flash 
--						(Bootup_Trigger, Bootup_Backup_Done).
--
-- 2. Set Verify_All and trigger the parameter update process 
--						(Update_Trigger, Update_Proc_Done).
--
-- 3. Clear Verify_All.
--
-- 4. Initialize the Curve_buffer (Set_CTBFREE_Sem).
--
-- 5. Preset the status bits for all of the curves (Preset_Curve_status).
--
-- 6. Run the curve interpolation routine EIGHT times 
--						(Curve_Intrp_Trigger, Crv_Intrp_Proc_Done).
-- 7. Signal that the DDS Module is ready.
-------------------------------------------------------------------------------

entity dds_startup_sequence is 
port( 
		clk					: in std_logic;
		nreset				: in std_logic;
		
		Enable_Proc_Sequence : in std_logic; -- forces re-initialization (resets state machine)
		
		Update_Proc_Done	: in std_logic; -- Parameter update has completed
		Crv_Req_Proc_Done	: in std_logic; -- Curve buffer process has completed
		Crv_Intrp_Proc_Done : in std_logic; -- Curve Interpolation process has completed
											-- and Block Erase has begun
		Bootup_Backup_Done	: in std_logic; -- A Flash store/restore has completed
		DDS_Setup_Done		: in std_logic; -- AD9910 DDS chip have been setup/initialized
		
		Setup_AD9910		: out std_logic; -- Trigger the AD9910, DDS setup/initialization
		Verify_All			: out std_logic; -- Forces update of all Parameters
		Set_CTBFREE_Sem		: out std_logic; -- Initialize the CBTFREE SEM
		Preset_Curve_Status : out std_logic; -- Forces all curves to be setup from buffers
		
		Update_Trigger		: out std_logic; -- Trigger the Parameter Update Process
		Curve_Intrp_Trigger : out std_logic; -- Trigger the Curve Interpolation Process
		Bootup_Trigger		: out std_logic; -- Trigger the Flash Restore Process
		Flash_Bus_Permit	: out std_logic; -- Signal Flash to continue the Store Process
		
		DDS_Ready			: out std_logic;
		Startup_State		: out std_logic_vector(7 downto 0)
	);
end dds_startup_sequence;
	
architecture main of dds_startup_sequence is	
-- Syncronized Signals ----------------------------------------------
	signal ad9910_setup_done : std_logic;

-- Start Delay Timer ------------------------------------------------
	signal sync_ena_seq		: std_logic_vector(1 downto 0);
	signal ena_proc_seq		: std_logic;
	signal start_seq  		: std_logic;
	signal start_timer		: std_logic_vector(15 downto 0);
	
-- Curve Interpolation Counter --------------------------------------
    signal intrp_cnt_ena	: std_logic;
    signal intrp_cnt_rst	: std_logic;
    signal intrp_count		: std_logic_vector(3 downto 0);
    
-- TIMER_STATE_MACHINE ----------------------------------------------
	type STATE_TYPE is
(init, do_AD9910_setup, boot_from_flash, wait_for_boot, update_all_params, init_crv_buffer, 
 preset_crv_status, do_crv_interp, inc_intrp_count, initialize_run, 
 is_dds_ready, dds_is_ready);
 		
 	signal CS, NS : STATE_TYPE;

--===================================================================
--			B E G I N
--===================================================================
begin	

SYNCRONIZE_ENABLE_PROCESS_SEQUENCE:
process(clk, nreset, Enable_Proc_Sequence, sync_ena_seq)
begin
	if   (nreset = '0') then sync_ena_seq <= (others => '0');
	elsif rising_edge(clk) then
		sync_ena_seq(0) <= Enable_Proc_Sequence;
		sync_ena_seq(1) <= sync_ena_seq(0);
	end if;
	
	ena_proc_seq <= sync_ena_seq(1);

end process;

  
SYNC_SETUP: led -- syncronize the signal DDS_Setup_Done.
port map(	sig_in  => DDS_Setup_Done, 
				clk     => clk, 
				clrn    => '1', 
				pls_out => ad9910_setup_done); 

--====================================================
COUNT_EIGHT_INTERPOLATIONS:
process(clk,nreset,intrp_cnt_ena, intrp_cnt_rst, intrp_count)
begin
	if   (nreset = '0')    then intrp_count <= (others => '0');
	elsif rising_edge(clk) then
		if   (intrp_cnt_ena = '1') then 	intrp_count <= intrp_count + 1;
		elsif(intrp_cnt_rst = '1') then 	intrp_count <= (others => '0');
		else										intrp_count <= intrp_count;
		end if;
	end if;
end process;
			
--====================================================
DELAY_START_TIMER:  -- delay at power up / boot up

process(clk, nreset, start_timer, ena_proc_seq)
begin
	if   (nreset = '0') OR (ena_proc_seq = '0') then
											start_timer <= (others => '0');
											start_seq  <= '0';
	elsif rising_edge(clk) then
		if   (start_timer < X"FFF0") then	
											start_timer <= start_timer + 1;
											start_seq  <= '0';
		else
											start_timer <= start_timer;
											start_seq  <= '1';
		end if;
	end if;
end process;

--====================================================
-- STATE MACHINE 
--**--   --**--   --**--   --**--   --**--   --**--   --**--  
SM_REG: process(clk, nreset, ena_proc_seq)
begin
	if(nreset = '0') OR (ena_proc_seq = '0') then CS <= init;
	elsif(rising_edge(clk))		then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process
(CS, Bootup_Backup_Done, intrp_count, start_seq, Update_Proc_Done,
 Crv_Req_Proc_Done, Crv_Intrp_Proc_Done, DDS_Setup_Done,ad9910_setup_done)

begin
	-- Defaults -------------------------------------
	Setup_Ad9910 		<= '0';
	Verify_All			<= '0';
	Set_CTBFREE_Sem		<= '0';
	Preset_Curve_Status <= '0';

	Update_Trigger		<= '0';
	Curve_Intrp_Trigger <= '0';
	Bootup_Trigger		<= '0';
	Flash_Bus_Permit	<= '0';
	
	intrp_cnt_ena		<= '0';
	intrp_cnt_rst		<= '0';
	
	DDS_Ready			<= '0'; -- held low until ready
	Startup_State 		<= X"FF";
	
	--------------------------------------------------
	case CS is
		when init =>									Startup_State <= "00000001";
			if   (start_seq = '1') then			NS <= boot_from_flash;
			else							NS <= init;
			end if;
	------------------------------------------------------------------
	-- POWER UP / BOOT UP / SET UP SECTION --------------------------		
		when boot_from_flash =>						NS <= wait_for_boot;
			Flash_Bus_Permit	<= '1';
			Bootup_Trigger		<= '1';
		
		when wait_for_boot =>						Startup_State <= "00000010";
			Flash_Bus_Permit	<= '1';
			if (Bootup_Backup_Done = '1') then 	NS <= update_all_params;
			else                   					NS <= wait_for_boot;
			end if;
		
		when update_all_params =>					Startup_State <= "00000100";
			Verify_All 		<= '1';
			Update_Trigger	<= '1';
			if (Update_Proc_Done = '1') then		NS <= init_crv_buffer;
			else											NS <= update_all_params;
			end if;
		
		when init_crv_buffer =>						Startup_State <= "00001000";
			Set_CTBFREE_Sem	<= '1';
			if (Crv_Req_Proc_Done = '1') then	NS <= preset_crv_status;
			else											NS <= init_crv_buffer;
			end if;
			
		when preset_crv_status =>					NS <= do_crv_interp;
			Preset_Curve_Status 	<= '1';
			intrp_cnt_rst			<= '1';
			
		when do_crv_interp =>						Startup_State <= "00010000";
			if (intrp_count < "1000") then
				Curve_Intrp_Trigger <= '1';
				if (Crv_Intrp_Proc_Done = '1') then	NS <= inc_intrp_count;
				else											NS <= do_crv_interp;
				end if;
			else												NS <= do_AD9910_setup;
			end if;
			
		when inc_intrp_count =>						NS <= do_crv_interp;
			intrp_cnt_ena 		<= '1'; -- increment the interpolation count
			
		------------------------------------------------------------------
		when do_AD9910_setup =>						NS <= is_dds_ready;
			Setup_AD9910 		<= '1';
			
		when is_dds_ready => 						Startup_State <= "00100000";
			-- check to see that the AD9910 DDS chips have been setup and initialized
			if   (ad9910_setup_done = '1') then		NS <= dds_is_ready;
			else									NS <= is_dds_ready;
			end if;
		
		when dds_is_ready =>						Startup_State <= X"F0";
														NS <= dds_is_ready;
			DDS_Ready			<= '1';
									
		when others =>              			NS <= init;
	end case;
end process;

end;